#!/usr/bin/env python
"""
Rotate linearly polarized data into Stokes' I,Q,U,V
"""
import aipy as a
import numpy as np
import optparse,sys,os

o = optparse.OptionParser()
o.set_usage('stokes_rotate.py *.uv')
o.set_description(__doc__)
opts,args = o.parse_args(sys.argv[1:])


for uvfile in args:
    
    infile = uvfile
    outfile = infile+'S'
    
    print infile,'-->',outfile
    if os.path.exists(outfile):
        print 'File exists, skipping'
        continue

    uv = a.pol.UV(uvfile)
    DD = {}
    for (uvw,t,bl),d,f in uv.all(raw=True):
        plzn = uv.read_pol()
        if not bl in DD.keys(): DD[bl] = {}
        if not t in DD[bl].keys(): DD[bl][t] = {}
        if not plzn in DD[bl][t].keys():
            DD[bl][t][plzn] = np.ma.array(d,mask=f)
    del(uv)
    
    for bl in DD:
        for t in DD[bl]:
            DD[bl][t] = a.pol.xy2stokes(DD[bl][t])
    
    def mfunc(uv,p,d,f):
        uvw,t,bl = p
        print uvi['pol']
        plzn = uvi.read_pol()
        print bl,plzn 
        if plzn == 'xx':
            uvo.write_pol('I')
            print '-->',uvo.read_pol()
            return p,DD[bl][t]['I'],f
        if plzn == 'xy': 
            uvo.write_pol('Q')
            print '-->',uvo.read_pol()
            return p,DD[bl][t]['Q'],f
        if plzn == 'yx': 
            uvo.write_pol('U')
            print '-->',uvo.read_pol()
            return p,DD[bl][t]['U'],f
        if plzn == 'yy': 
            uvo.write_pol('V')
            print '-->',uvo.read_pol()
            return p,DD[bl][t]['V'],f
    
    uvi = a.pol.UV(infile)
    uvo = a.pol.UV(outfile,status='new')
    uvo.init_from_uv(uvi)
    uvo.pipe(uvi,raw=True,mfunc=mfunc,append2hist='XY --> STOKES \n')



#!/usr/bin/env python

"""
Get a first-pass gain calibration by enforcing amplitude closure.
"""

import aipy as a
import numpy as np
from matplotlib import pylab as pl
import optparse,sys,os
from time import time

o = optparse.OptionParser()
o.set_description(__doc__)
o.set_usage('amp_close.py [options] *.uv')
o.add_option('-C','--cal',dest='cal',default=None,help='Cal file to use')
o.add_option('-f','--flag',dest='flag',type='str',help='Antennae to flag. Ex. 1,2,3')
o.add_option('-p','--plots',dest='plots',default=None,action='store_true',help='Plots? [default=no]')
o.add_option('-s','--src',dest='src',default='None',help='Flux Calibration Source [default flux = 1]')
o.add_option('--minuv',dest='minuv',type='float',default=30,help='Minimum baseline length to consider [default=30lambda]')
o.add_option('--wgt',dest='wgt',type='str',default='equal',help='Weighting scheme, options are equal, radial, var, snr [default=equal]')
opts,args = o.parse_args(sys.argv[1:])

t0 = time()

#Parse Arguments:
FlagAnt = []
if opts.flag:
    for antstr in opts.flag.split(','):
        FlagAnt.append(int(antstr))
def uvlen(A):
    return np.sqrt(np.dot(A,A))
aa = a.cal.get_aa(opts.cal,np.array([0.15]))

#Find source Flux:
if opts.src == None: flx = 1.
else: 
    srclist,cutoff,catalogs = a.scripting.parse_srcs(opts.src,'helm,misc')
    cat = a.cal.get_catalog(opts.cal,srclist,cutoff,catalogs)
    cat.compute(aa)
    flx_list = cat.get_jys()
    flx = 0.
    print "Setting Jansky levels to the following sources:"
    for i,src in enumerate(srclist):
        print ' '*5,src,flx_list[i].squeeze(),'Jys'
        flx += flx_list[i].squeeze()

#Read in Data:
DD,cnt,bl_len,AntNos = {},{},{},[]

for uvfile in args:
    print 'Reading',uvfile
    uv = a.miriad.UV(uvfile)
    for (uvw,t,(i,j)),d,f in uv.all(raw=True):
        pol = a.miriad.pol2str[uv['pol']]
        #Selection Criteria:
        if i == j: continue
        if np.sum(f) == uv['nchan']: continue
        if pol[0] != pol[1]: continue
        if i in FlagAnt or j in FlagAnt: continue
        if not (i,j) in bl_len.keys(): bl_len[(i,j)] = uvlen(aa.get_baseline(i,j))
        if bl_len[(i,j)]*0.15 <= opts.minuv: continue
        #Populate arrays:
        if not pol in DD.keys(): 
            DD[pol] = {}
            cnt[pol] = {}
        if not (i,j) in DD[pol].keys(): 
            DD[pol][(i,j)] = np.zeros(uv['nchan'],dtype=np.complex)
            cnt[pol][(i,j)] = np.zeros(uv['nchan'])
        DD[pol][(i,j)] += np.abs(d)
        cnt[pol][(i,j)] += np.where(f,0,1)
        if not i in AntNos: AntNos.append(i)
        if not j in AntNos: AntNos.append(j)
pols = DD.keys()
AntNos.sort()
pols.sort()
dnu = uv['sdf']
del(uv)

Npol,Nant = len(DD.keys()),len(AntNos)
Nbl = int(Nant*(Nant-1)/2)

#Integrate and all that shit...
G_ij,W_ij = {},{}
for pol in pols:
    G_ij[pol] = {}
    W_ij[pol] = {}
    for bl in DD[pol]:
        DD[pol][bl] = np.where(cnt[pol][bl] == 0, 0, DD[pol][bl]/cnt[pol][bl])
        G_ij[pol][bl] = np.log(np.mean(DD[pol][bl]))
        if opts.wgt == 'equal' : W_ij[pol][bl] = 1.
        if opts.wgt == 'radial': W_ij[pol][bl] = bl_len[bl]
        if opts.wgt == 'var'   : W_ij[pol][bl] = 1./np.log(np.var(DD[pol][bl]))
        if opts.wgt == 'snr'   : W_ij[pol][bl] = np.log(np.mean(DD[pol][bl]))/np.var(DD[pol][bl])

#Do the linear algebra
G = {}
for pol in pols:
    A = np.zeros((Nbl,Nant))
    W = np.zeros((Nbl,Nbl))
    b = np.zeros(Nbl)
    
    #populatin the countryside...
    index = 0
    for i,ant1 in enumerate(AntNos):
        for j,ant2 in enumerate(AntNos):
            if not (ant1,ant2) in G_ij[pol].keys(): continue
            b[index] = G_ij[pol][(ant1,ant2)]
            A[index,i],A[index,j] = 1.,1.
            W[index,index] = W_ij[pol][(ant1,ant2)]
            index += 1
    W /= np.sum(W)

    G[pol] = 10**np.dot(np.linalg.inv(np.dot(np.dot(A.T,W),A)),np.dot(np.dot(A.T,W),b))
    G[pol] /= flx*np.mean(G[pol])

print "'amps' : {"
for i,ant in enumerate(AntNos):
    ampstr = str(ant)+' : {'
    for pol in pols:
        ampstr += "'"+pol[0]+"' : "+str(G[pol][i])
        if pol != pols[-1]: ampstr += ', '
    ampstr += ' },'
    print ampstr
print '},'

t1 = time()
print 'Computation time =',t1-t0,'s'

if opts.plots:
    print 'Generating plots... this may take a while...'
    AntNos = np.array(AntNos)
    figcnt = 0

    for pol in pols:
        pl.figure(figcnt)
        DDD = np.zeros((Nant,Nant))
        for i,ant1 in enumerate(AntNos):
            for j,ant2 in enumerate(AntNos):
                if not (ant1,ant2) in G_ij[pol].keys(): continue
                DDD[i,j] = 10**G_ij[pol][(ant1,ant2)]/flx
        DDD = np.ma.array(DDD,mask=np.where(DDD==0,1,0))
        pl.imshow(DDD,aspect='auto',interpolation='nearest')
        pl.colorbar()
        pl.title('Gains by baseline, polarization %s'%pol[0])
        pl.draw()
        figcnt += 1
    
    if opts.wgt != 'equal':
        for pol in pols:
            pl.figure(figcnt)
            WWW = np.zeros((Nant,Nant))
            for i,ant1 in enumerate(AntNos):
                for j,ant2 in enumerate(AntNos):
                    if not (ant1,ant2) in W_ij[pol].keys(): continue
                    WWW[i,j] = W_ij[pol][(ant1,ant2)]
            WWW = np.ma.array(WWW/np.sum(WWW),mask=np.where(WWW==0,1,0))
            pl.imshow(WWW,aspect='auto',interpolation='nearest')
            pl.colorbar()
            pl.title('Weights for each baseline, polarization %s'%pol[0])
            pl.draw()
            figcnt += 1 

    pl.figure(figcnt)
    for pol in pols:
        pl.plot(AntNos,G[pol],'o',label=pol[0])
    pl.xlabel('Antenna Number')
    pl.ylabel('Gain (units)')
    pl.legend()
    pl.draw()
    figcnt += 1

    pl.show()




'''The 6C (Sixth Cambridge) Catalog.
Data files are in tab-separated format from Vizier.
To download in the correct format, open a catalog online in Vizier,
select'Tab-Separated-Values' as the Output layout in the drop-down box, set
the maximum entries to 'unlimited', and click 'Sexagesimal' under the box
for 'Target Name or Position'.  Submit the query, and copy the output to a
txt file.  Copy files to "6c1.txt", "6c2.txt", "6c3.txt", "6c4.txt", "6c5_1.txt",
and "6c5_2.txt" (for the 5 fields) in the _src directory of your AIPY
installation.'''
import aipy as a, numpy as np, os

class SixCCatalog(a.fit.SrcCatalog):
    def fromfile(self,filename):
        f = open(filename)
        addsrcs = []
        for L in [L for L in f.readlines() if not L.startswith('#')]:
            text = L.split('\t')
            if len(text) <= 4: continue
            try: int(text[0][0])
            except(ValueError): continue
            ra = text[0].replace(' ',':') 
            dec = text[1].replace(' ',':') 
            name = '%s_%s' % (ra,dec)
            jys = float(text[4])
            addsrcs.append(a.fit.RadioFixedBody(ra, dec, name=name,
                jys=jys, index=0, mfreq=.151))
        self.add_srcs(addsrcs)

SIXCFILES = [os.path.dirname(__file__) + os.sep + cat 
    for cat in ['6c1.txt','6c2.txt','6c3.txt','6c4.txt','6c5_1.txt','6c5_2.txt']]
_sixccat = None

def get_srcs(srcs=None, cutoff=None):
    global _sixccat
    if _sixccat is None:
        _sixccat = SixCCatalog()
        for f in SIXCFILES: _sixccat.fromfile(f)
    if srcs is None:
        if cutoff is None: srcs = _sixccat.keys()
        else:
            cut, fq = cutoff
            fq = np.array([fq])
            for s in _sixccat.keys(): _sixccat[s].update_jys(fq)
            srcs = [s for s in _sixccat.keys() if _sixccat[s].jys[0] > cut]
    srclist = []
    for s in srcs:
        try: srclist.append(_sixccat[s])
        except(KeyError): pass
    return srclist



import numpy as np
from aipy._cephes import i0
from _dsp import *

WINDOW_FUNC = {
    'blackman': lambda x,L: .42-.5*np.cos(2*np.pi*x/(L-1))+.08*np.cos(4*np.pi*x/(L-1)),
    'blackman-harris': lambda x,L: .35875 - .48829*np.cos(2*np.pi*x/(L-1)) + .14128*np.cos(4*np.pi*x/(L-1)) - .01168*np.cos(6*np.pi*x/(L-1)),
    'gaussian0.4': lambda x,L: np.exp(-0.5 * ((x - (L-1)/2)/(0.4 * (L-1)/2))**2),
    'kaiser2': lambda x,L: i0(np.pi * 2 * np.sqrt(1-(2*x/(L-1) - 1)**2)) / i0(np.pi * 2),
    'kaiser3': lambda x,L: i0(np.pi * 3 * np.sqrt(1-(2*x/(L-1) - 1)**2)) / i0(np.pi * 3),
    'hamming': lambda x,L: .54 - .46 * np.cos(2*np.pi*x/(L-1)),
    'hanning': lambda x,L: .5 - .5 * np.cos(2*np.pi*x/(L-1)),
    'parzen': lambda x,L: 1 - np.abs(L/2. - x) / (L/2.),
    'none': lambda x,L: 1,
}

def gen_window(L, window='hamming'):
    '''Return the specified window (see WINDOW_FUNC) for a length L.'''
    return np.fromfunction(lambda x: WINDOW_FUNC[window](x,L), (L,))



"""
The PAPER catalog here is derived by experimental methods in either facets or 
healpix maps.
"""

import aipy as a, numpy as np, os
    
class PAPERCatalog(a.fit.SrcCatalog):
    def fromfile(self, filename):
        f = open(filename)
        addsrcs = []
        for L in [L for L in f.readlines() if not L.startswith('#')]:
            text = L.split('\t')
            if len(text) <= 3: continue
            try: int(text[0].strip()[0])
            except(ValueError): continue
            ra = text[1]
            dec = text[2]
            name = text[3].strip()
            jys = float(text[4])
            addsrcs.append(a.fit.RadioFixedBody(ra, dec, name=name,
                jys=jys, index=0, mfreq=.150))
        self.add_srcs(addsrcs)

PAPERFILE = os.path.dirname(__file__) + os.sep + 'paper.txt'
_papercat = None

def get_srcs(srcs=None, cutoff=None):
    global _papercat
    if _papercat is None:
        _papercat = PAPERCatalog()
        _papercat.fromfile(PAPERFILE)
    if srcs is None:
        if cutoff is None: srcs = _papercat.keys()
        else:
            cut, fq = cutoff
            fq = np.array([fq])
            for s in _papercat.keys(): _papercat[s].update_jys(fq)
            srcs = [s for s in _papercat.keys() if _papercat[s].jys[0] > cut]

    srclist = []
    for s in srcs:
        try: srclist.append(_papercat[s])
        except(KeyError): pass
    return srclist



# -*- coding: utf-8 -*-
import unittest, aipy._healpix as h, numpy as n

class TestHealpix(unittest.TestCase):
    def setUp(self):
        self.hpb = h.HealpixBase()
    def test_order(self):
        """Test HEALpix order funtional attribute"""
        self.assertEqual(self.hpb.order(), -1)
    def test_nside(self):
        """Test HEALpix nside funtional attribute"""
        self.assertEqual(self.hpb.nside(), 0)
    def test_npix(self):
        """Test HEALpix npix funtional attribute"""
        self.assertEqual(self.hpb.npix(), 12*self.hpb.nside()**2)
    def test_scheme(self):
        """Test HEALpix scheme funtional attribute"""
        self.assertEqual(self.hpb.scheme(), 'RING')
    def test_npix2nside(self):
        """Test HEALpix npix2nside funtional attribute"""
        self.assertEqual(self.hpb.npix2nside(12*2**12), 2**6)

if False:
  class TestMemLeaks(unittest.TestCase):
    def setUp(self):
        self.hpb = h.HealpixBase(nside=256)
    def test_create(self):
        while True: hpb = h.HealpixBase(nside=256)
    def test_crd2pix(self):
        one = 0.99
        x = n.arange(-one, one, .01, dtype=n.double)
        y = n.zeros_like(x)
        z = n.sqrt(1 - x**2 - y**2)
        while True: crds = self.hpb.crd2px(x, y, z)

class TestSuite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the aipy._healpix unit tests."""

    def __init__(self):
        unittest.TestSuite.__init__(self)

        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(TestHealpix))
        #self.addTests(loader.loadTestsFromTestCase(TestMemLeaks))

if __name__ == '__main__':
    unittest.main()

#hp = aipy._healpix.HealpixBase(2**6, 'NEST')
#print hp.order(), hp.nside(), hp.npix(), hp.scheme()
#px = numpy.arange(hp.npix())
#px = hp.nest_ring_conv(px, 'RING')
#hp.set_nside_scheme(hp.nside(), 'RING')
#px = numpy.arange(10)
#x,y,z = hp.px2crd(px, ncrd=3)
#th,phi = hp.px2crd(px, ncrd=2)
#assert(numpy.all(px == hp.crd2px(x,y,z, interpolate=0)))
#px2,wgt = hp.crd2px(x,y,z, interpolate=1)
#z[-1] = numpy.Inf
#try: hp.crd2px(x,y,z[:-1])
#except(RuntimeError): pass
#assert(numpy.all(px == hp.crd2px(th, phi, interpolate=0)))
#px2,wgt = hp.crd2px(th, phi, interpolate=1)



'''The Culgoora Catalog.
Data files are in tab-separated format from Vizier.
To download in the correct format, open a catalog online in Vizier,
select'Tab-Separated-Values' as the Output layout in the drop-down box, set
the maximum entries to 'unlimited', and click 'Sexagesimal' under the box
for 'Target Name or Position'.  Submit the query, and copy the output to a
txt file.  Copy this file to "culgoora.txt" in the _src directory of your AIPY
installation.'''
import aipy as a, numpy as np, os

class CulgooraCatalog(a.fit.SrcCatalog):
    metadata = {}
    def get_metadata(self):
        '''Return the actual 80 MHz and 160 MHz measurements used to compute
        the spectral index in the Culgoora catalog. Returns dictionary with
        source names linked to (S80, S160, index) triplets.  In cases where 
        measurements are unavailable, values are entered as None.'''
        return self.metadata
    def fromfile(self,filename):
        '''Parses a tab-delimited text file for the Culgoora array.  When both
        80 MHz and 160 MHz measurements are available, catalog entries use
        mfreq=.160 and include a computed spectral index.  When only one
        measurement is available, mfreq is set for that measurement, and index
        is set to 0.'''
        f = open(filename)
        addsrcs = []
        for L in [L for L in f.readlines() if not L.startswith('#')]:
            text = L.split('\t')
            if len(text) <= 4: continue
            try: int(text[0][0])
            except(ValueError): continue
            ra = text[0].replace(' ',':')
            dec = text[1].replace(' ',':')
            name = text[2].strip()
            try: jys160 = float(text[7])
            except(ValueError): jys160 = None
            try: jys080 = float(text[9])
            except(ValueError): jys080 = None
            try: index = float(text[10])
            except(ValueError): index = None
            # Catch flagged flux-densities
            if index is None and (not jys080 is None and not jys160 is None):
                if text[6].startswith('<'): jys160 = None
                if text[8].startswith('<'): jys080 = None
            self.metadata[name] = (jys080, jys160, index)
            if not index is None:
                addsrcs.append(a.fit.RadioFixedBody(ra, dec, name=name,
                    jys=jys160, index=index, mfreq=0.160))
            elif not jys160 is None:
                addsrcs.append(a.fit.RadioFixedBody(ra, dec, name=name,
                    jys=jys160, index=0, mfreq=0.160))
            else:
                addsrcs.append(a.fit.RadioFixedBody(ra, dec, name=name,
                    jys=jys080, index=0, mfreq=0.080))
        self.add_srcs(addsrcs)

CULGOORAFILE = os.path.dirname(__file__) + os.sep + 'culgoora.txt'
_culgooracat = None

def get_srcs(srcs=None, cutoff=None):
    global _culgooracat
    if _culgooracat is None:
        _culgooracat = CulgooraCatalog()
        _culgooracat.fromfile(CULGOORAFILE)
    if srcs is None:
        if cutoff is None: srcs = _culgooracat.keys()
        else:
            cut, fq = cutoff
            fq = np.array([fq])
            for s in _culgooracat.keys(): _culgooracat[s].update_jys(fq)
            srcs = [s for s in _culgooracat.keys() if _culgooracat[s].jys[0] > cut]
    srclist = []
    for s in srcs:
        try: srclist.append(_culgooracat[s])
        except(KeyError): pass
    return srclist



# -*- coding: utf-8 -*-
import unittest
import aipy.amp as amp, numpy as n

class TestRadioBody(unittest.TestCase):
    def setUp(self):
        self.fqs = n.arange(.1,.2,.01)
        bm = amp.Beam(self.fqs)
        ant0 = amp.Antenna(0,0,0,bm)
        self.aa = amp.AntennaArray(('0','0'), [ant0])
    def test_attributes(self):
        """Test aipy.amp.RadioFixedBody attributes"""
        s = amp.RadioFixedBody('0:00', '0:00',
            jys=100, index=-2, mfreq=.1, name='src1')
        self.assertEqual(s._jys, 100)
        self.assertEqual(s.index, -2)
        s.compute(self.aa)
        self.assertTrue(n.all(s.get_jys() == 100 * (self.fqs / .1)**-2))

class TestBeam(unittest.TestCase):
    def setUp(self):
        self.fqs = n.arange(.1,.2,.01)
        self.bm = amp.Beam(self.fqs)
    def test_response(self):
        """Test retrieving aipy.amp.Beam response"""
        xyz = (0,0,1)
        self.assertTrue(n.all(self.bm.response(xyz) == n.ones_like(self.fqs)))
        x = n.array([0, .1, .2])
        y = n.array([.1, .2, 0])
        z = n.array([.2, 0, .1])
        xyz = (x,y,z)
        self.assertTrue(n.all(self.bm.response(xyz) == \
            n.ones((self.fqs.size,3))))
        self.bm.select_chans([0,1,2])
        self.assertTrue(n.all(self.bm.response(xyz) == n.ones((3,3))))

class TestBeam2DGaussian(unittest.TestCase):
    def setUp(self):
        self.fqs = n.arange(.1,.2,.01)
        self.bm = amp.Beam2DGaussian(self.fqs, .05, .025)
    def test_response(self):
        """Test retrieving a 2D Gaussian beam response"""
        xyz = (0,0,1)
        self.assertTrue(n.all(self.bm.response(xyz) == n.ones_like(self.fqs)))
        x = n.array([0, .05,   0])
        y = n.array([0,   0, .05])
        z = n.array([1,   1,   1])
        xyz = (x,y,z)
        resp = self.bm.response(xyz)
        self.assertEqual(resp.shape, (self.fqs.size,3))
        ans = n.sqrt(n.array([1., n.exp(-0.5), n.exp(-2)]))
        ans.shape = (1,3)
        self.bm.select_chans([0])
        resp = self.bm.response(xyz)
        self.assertTrue(n.all(n.round(resp - ans, 3) == 0))

class TestBeamAlm(unittest.TestCase):
    def setUp(self):
        self.fqs = n.arange(.1,.2,.01)
        self.coeffs = {0:n.array([1.,0,0], dtype=n.complex)}
        self.bm = amp.BeamAlm(self.fqs, lmax=1, mmax=1, deg=0, coeffs=self.coeffs)
    def test_init_update(self):
        self.assertTrue(n.allclose(self.bm.hmap[0].map, 1./n.sqrt(4*n.pi)))
    def test_response(self):
        x = n.array([0, .05,   0])
        y = n.array([0,   0, .05])
        z = n.array([1,   1,   1])
        xyz = (x,y,z)
        resp = self.bm.response(xyz)
        self.assertEqual(resp.shape, (self.fqs.size,3))
        ans = n.ones(3, dtype=n.complex) / n.sqrt(4*n.pi)
        ans.shape = (1,3)
        self.bm.select_chans([0])
        resp = self.bm.response(xyz)
        self.assertTrue(n.all(n.round(resp - ans, 3) == 0))

class TestAntenna(unittest.TestCase):
    def setUp(self):
        self.fqs = n.arange(.1,.2,.01)
        bm = amp.Beam2DGaussian(self.fqs, .05, .025)
        self.ant = amp.Antenna(0,0,0, beam=bm)
    def test_passband(self):
        """Test the Antenna passband"""
        pb = self.ant.passband()
        self.assertTrue(n.all(pb == n.ones_like(self.fqs)))
        self.ant.select_chans([0,1,2])
        pb = self.ant.passband()
        self.assertEqual(pb.shape, (3,))
    def test_bm_response(self):
        """Test the Antenna beam response"""
        xyz = (.05,0,1)
        self.ant.select_chans([0])
        # resp = self.ant.bm_response(xyz, pol='x')
        # self.assertAlmostEqual(resp, n.sqrt(n.exp(-0.5)), 3)
        # resp = self.ant.bm_response(xyz, pol='y')
        # self.assertAlmostEqual(resp, n.sqrt(n.exp(-2)), 3)

#class TestMemLeaks(unittest.TestCase):
#    def test_antenna_create(self):
#        freqs = n.arange(.1,.2,.001)
#        beam = amp.Beam(freqs)
#        while True: ant = amp.Antenna(0,0,0,beam, pointing=(0,n.pi/2,.1))
#    def test_aa_create(self):
#        freqs = n.arange(.1,.2,.001)
#        beam = amp.Beam(freqs)
#        ants = [amp.Antenna(0,0,0,beam) for i in range(100)]
#        while True: aa = amp.AntennaArray(('0:00','0:00'), ants)
        
class TestSuite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the aipy.amp unit tests."""

    def __init__(self):
        unittest.TestSuite.__init__(self)

        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(TestRadioBody))
        self.addTests(loader.loadTestsFromTestCase(TestBeam))
        self.addTests(loader.loadTestsFromTestCase(TestBeam2DGaussian))
        self.addTests(loader.loadTestsFromTestCase(TestAntenna))
        #self.addTests(loader.loadTestsFromTestCase(TestMemLeaks))

if __name__ == '__main__':
    unittest.main()



"""
Module containing utilities (like parsing of certain command-line arguments) 
for writing scripts.
"""

import miriad, fit, src, numpy as np, re,phs

def add_standard_options(optparser, ant=False, pol=False, chan=False, 
        cal=False, src=False, prms=False, dec=False, cmap=False, 
        max=False, drng=False):
    """Add standard command-line options to an optparse.OptionParser() on an 
    opt in basis (i.e. specify =True for each option to be added)."""
    if ant: optparser.add_option ('-a', '--ant', dest='ant', default='cross',
         help='Select ant_pol/baselines to include.  Examples: all (all baselines) auto (of active baselines, only i=j) cross (only i!=j) 0,1,2 (any baseline involving listed ants) 0_2,0_3 (only listed baselines) "(0,1)_(2,3)" (same as 0_2,0_3,1_2,2_3. Quotes help bash deal with parentheses) "(-0,1)_(2,-3)" (exclude 0_2,0_3,1_3 include 1_2).  Default is "cross". Select pol by adding appropriate x or y eg 5x_6y.')
    if pol: optparser.add_option('-p', '--pol', dest='pol', default=-1,
        help='Choose polarization (xx, yy, xy, yx) to include.')
    if chan: optparser.add_option('-c', '--chan', dest='chan', default='all',
        help='Select channels (after any delay/delay-rate transforms) to include.  Examples: all (all channels), 0_10 (channels from 0 to 10, including 0 and 10) 0_10_2 (channels from 0 to 10, counting by 2), 0,10,20_30 (mix of individual channels and ranges).  Default is "all".')
    if cal: optparser.add_option('-C', '--cal', dest='cal', 
        help='Use specified <cal>.py for calibration information.')
    if src:
        optparser.add_option('-s', '--src', dest='src',
            help='Phase centers/source catalog entries to use.  Options are "all", "<src_name1>,...", or "<ra XX[:XX:xx]>_<dec XX[:XX:xx]>".')
        optparser.add_option('--cat', dest='cat', default='helm,misc',
            help='A comma-delimited list of catalogs from which sources are to be drawn.  Default is "helm,misc".  Other available catalogs are listed under aipy._src.  Some catalogs may require a separate data file to be downloaded and installed.')
    if prms: optparser.add_option('-P', '--prms', dest='prms',
        help='Parameters (for fitting, usually), can be specified as can be: "obj=prm", "obj=prm/val", "obj=prm/val/sig", "(obj1/obj2)=prm/(val1/val2)/sig", "obj=(prm1/prm2)/val/(sig1/sig2)", comma separated versions of the above, and so on.')
    if dec:
        optparser.add_option('-x', '--decimate', dest='decimate', 
            default=1, type='int',
            help='Use only every Nth integration.  Default is 1.')
        optparser.add_option('--dphs', dest='decphs', 
            default=0, type='int',
            help='Offset to use when decimating (i.e. start counting integrations at this number for the purpose of decimation).  Default is 0.')
    if cmap: optparser.add_option('--cmap', dest='cmap', default='jet',
        help='Colormap for plotting.  Can be gist_earth, gist_heat, gist_stern, gist_yarg, hot, cool, gray, bone, spectral, copper, jet to name a few.  For a more complete list, see pylab.cm.datad.keys().  Default is jet.')
    if max: optparser.add_option( '--max',dest='max',type='float',default=None,
    help='Manually set the maximum color level, in units matching plotting mode.  Default max(data).')
    if drng:
        optparser.add_option('--drng', dest='drng', type='float', default=None,
    help="Dynamic range in color of image, in units matching plotting mode.  Default max(data)-min(data).")

ant_re = r'(\(((-?\d+[xy]?,?)+)\)|-?\d+[xy]?)'
bl_re = '(^(%s_%s|%s),?)' % (ant_re, ant_re, ant_re)
def parse_ants(ant_str, nants):
    """Generate list of (baseline, include, pol) tuples based on parsing of the
    string associated with the 'ants' command-line option."""
    rv,cnt = [], 0
    while cnt < len(ant_str):
        m = re.search(bl_re, ant_str[cnt:])
        if m is None:
            if ant_str[cnt:].startswith('all'): rv = []
            elif ant_str[cnt:].startswith('auto'): rv.append(('auto',1,-1))
            elif ant_str[cnt:].startswith('cross'): rv.append(('auto',0,-1))
            else: raise ValueError('Unparsible ant argument "%s"' % ant_str)
            c = ant_str[cnt:].find(',')
            if c >= 0: cnt += c + 1
            else: cnt = len(ant_str)
        else:
            m = m.groups()
            cnt += len(m[0])
            if m[2] is None:
                ais = [m[8]]
                ajs = range(nants)
            else:
                if m[3] is None: ais = [m[2]]
                else: ais = m[3].split(',')
                if m[6] is None: ajs = [m[5]]
                else: ajs = m[6].split(',')
            for i in ais:
                if type(i) == str and i.startswith('-'):
                     i = i[1:] #nibble the - off the string
                     include_i = 0 
                else: include_i = 1
                for j in ajs:
                    include = None
                    if type(j) == str and j.startswith('-'):
                        j = j[1:]
                        include_j = 0
                    else: include_j = 1
                    include = int(include_i and include_j)
                    pol = None
                    i,j = str(i),str(j)
                    if not i.isdigit():
                        ai = re.search(r'(\d+)([x,y])',i).groups()
                    if not j.isdigit():
                        aj = re.search(r'(\d+)([x,y])',j).groups()
                    if i.isdigit() and not j.isdigit():
                        pol = ['x'+aj[1],'y'+aj[1]]
                        ai = [i,'']
                    elif not i.isdigit() and j.isdigit():
                        pol = [ai[1]+'x',ai[1]+'y']
                        aj = [j,'']
                    elif not i.isdigit() and not j.isdigit():
                        pol = [ai[1]+aj[1]]
                    if not pol is None:
                        bl = miriad.ij2bl(abs(int(ai[0])),abs(int(aj[0])))
                        for p in pol:
                            rv.append((bl,include,p))
                    else: 
                        bl = miriad.ij2bl(abs(int(i)),abs(int(j)))
                        rv.append((bl,include,-1))
    return rv

def uv_selector(uv, ants=-1, pol_str=-1):
    """Call uv.select with appropriate options based on string argument for
    antennas (can be 'all', 'auto', 'cross', '0,1,2', or '0_1,0_2') and
    string for polarization ('xx','yy','xy','yx')."""
    if ants != -1:
        if type(ants) == str: ants = parse_ants(ants, uv['nants'])
        for cnt,(bl,include,pol) in enumerate(ants):
            if cnt > 0:
                if include: uv.select('or',-1,-1)
                else: uv.select('and',-1,-1)
            if pol == -1: pol = pol_str # default to explicit pol parameter
            if bl == 'auto': uv.select('auto', 0, 0, include=include)
            else:
                i,j = miriad.bl2ij(bl)
                uv.select('antennae', i, j, include=include)
            if pol != -1:
                for p in pol.split(','):
                    polopt = miriad.str2pol[p]
                    uv.select('polarization', polopt, 0)
    elif pol_str != -1:
        for p in pol_str.split(','):
            polopt = miriad.str2pol[p]
            uv.select('polarization', polopt, 0)


def parse_chans(chan_str, nchan, concat=True):
    """Return array of active channels based on number of channels and
    string argument for chans (all, 20_30, or 55,56,57, or 20_30,31,32).
    Channel ranges include endpoints (i.e. 20_30 includes both 20 and 30)."""
    if chan_str.startswith('all'): chanopt = [np.arange(nchan)]
    else:
        chanopt = []
        for co in chan_str.split(','):
            co = map(int, co.split('_'))
            assert(len(co) in [1,2,3])
            if len(co) == 1: chanopt.append(np.array(co))
            elif len(co) == 2: chanopt.append(np.arange(co[0],co[1]+1))
            else: chanopt.append(np.arange(co[0],co[1]+1,co[2]))
    if concat: return np.concatenate(chanopt)
    return chanopt

def parse_srcs(src_str, cat_str):
    """Return (src_list,flux_cutoff,catalogs) based on string argument for src and
    cat.  Can be "all", "<src_name1>,...", "<ra XX[:XX:xx]>_<dec XX[:XX:xx]>", or
    "val/freq" (sources with Jy flux density above val at freq in GHz)."""
    cats = cat_str.split(',')
    if src_str.startswith('all'): return None, None, cats
    if src_str.find('/') != -1:
        cutoff = map(float, src_str.split('/'))
        return None, cutoff, cats
    src_opt = src_str.split(',')
    for i, s in enumerate(src_opt):
        radec = s.split('_')
        try:
            assert(len(radec) == 2)
            ra,dec = radec
            # Check that this is really an ra_dec pair
            for piece in ra.split(':'): piece = float(piece)
            for piece in dec.split(':'): piece = float(piece)
            src_opt[i] = fit.RadioFixedBody(ra,dec, name=s)
        except(AssertionError,ValueError): continue
    return src_opt, None, cats

name = r'([^\(/,\)=]+)'
grp = r'(%s|\((%s(/%s)*)\))' % tuple([name]*3)
prm = r'(%s=%s(/(%s)?(/%s)?)?)' % tuple([grp]*4)
prm_rgx = re.compile(prm)
def parse_prms(prm_str):
    """Return a dict of the form: {'obj': {'prm':(val,sig),...}...} where
    val is a starting value and sig is a known error associated with that 
    start value.  Both default to None if a value is not provided.  The
    string to be parsed can be: "obj=prm", "obj=prm/val", 
    "obj=prm/val/sig", "(obj1/obj2)=prm/(val1/val2)/sig", 
    "obj=(prm1/prm2)/val/(sig1/sig2)", comma separated versions of the above,
    and so on."""
    prms = {}
    for prm in prm_str.split(','):
        m = prm_rgx.match(prm)
        g = m.groups()
        if g[2]: obj = [g[2]]
        else: obj = g[3].split('/')
        if g[8]: plist = [g[8]]
        else: plist = g[9].split('/')
        if g[16]: ival = [float(g[16])]
        elif g[17]: ival = map(float, g[17].split('/'))
        else: ival = [None]
        if g[23]: sval = [float(g[23])]
        elif g[24]: sval = map(float, g[24].split('/'))
        else: sval = [None]
        if len(obj) != 1:
            if len(plist) != 1:
                assert(len(ival) == 1 and len(sval) == 1)
                ival = [ival * len(plist)] * len(obj)
                sval = [sval * len(plist)] * len(obj)
            else:
                if len(ival) == 1: ival = ival * len(obj)
                if len(sval) == 1: sval = sval * len(obj)
                assert(len(ival) == len(obj) and len(sval) == len(obj))
                ival = [[i] for i in ival]
                sval = [[i] for i in sval]
        else:
            if len(plist) != 1:
                if len(ival) == 1: ival = ival * len(plist)
                if len(sval) == 1: sval = sval * len(plist)
                assert(len(ival) == len(plist) and len(sval) == len(plist))
            else:
                assert(len(ival) == 1 and len(sval) == 1)
            ival = [ival]
            sval = [sval]
        for o,il,sl in zip(obj,ival,sval):
            if not prms.has_key(o): prms[o] = {}
            for p,i,s in zip(plist,il,sl):
                prms[o][p] = (i,s)
    return prms

def get_null_aa():
   return phs.AntennaArray([0,0],
       [phs.Antenna(0,0,0,phs.Beam(np.array([0.15])))])



# -*- coding: utf-8 -*-
import sys
import unittest
import timeit

class TestSpeed(unittest.TestCase):
    def test_phsspeed(self):
        """Test the speed of the aipy.phs module"""
        setup = '''
import numpy as n, aipy as a
freqs = n.arange(.1,.2,.0001)
beam = a.fit.Beam(freqs)
ants = [
    a.fit.Antenna(  0,   0,  0, beam),
    a.fit.Antenna(  0, 100,  0, beam),
    a.fit.Antenna(100,   0,  0, beam),
    a.fit.Antenna(100, 100,  0, beam),
]
aa = a.fit.AntennaArray(('45:00','90:00'), ants)
aa.set_jultime(2455400.1)
s_eqs = n.array([[0,1,0]]*100).transpose()
'''
        expr = '''
aa.gen_phs(s_eqs,0,1,)
'''
        t = timeit.Timer(expr, setup=setup)
        sys.stderr.write("%.3f ms ... " % ((t.timeit(number=10)/ 10) * 1e3))

class TestSuite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the aipy.phs unit tests."""

    def __init__(self):
        unittest.TestSuite.__init__(self)

        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(TestSpeed))

if __name__ == '__main__':
    unittest.main()



"""
A module implementing various techniques for deconvolving an image by a
kernel.  Currently implemented are Clean, Least-Squares, Maximum Entropy,
and Annealing.  Standard parameters to these functions are:
im = image to be deconvolved.
ker = kernel to deconvolve by (must be same size as im).
mdl = a priori model of what the deconvolved image should look like.
maxiter = maximum number of iterations performed before terminating.
tol = termination criterion, lower being more optimized.
verbose =  print info on how things are progressing.
lower = lower bound of pixel values in deconvolved image
upper = upper bound of pixel values in deconvolved image
"""

import numpy as np
import sys
from . import _deconv

# Find smallest representable # > 0 for setting clip level
lo_clip_lev = np.finfo(np.float).tiny 

def clean(im, ker, mdl=None, area=None, gain=.1, maxiter=10000, tol=1e-3, 
        stop_if_div=True, verbose=False, pos_def=False):
    """This standard Hoegbom clean deconvolution algorithm operates on the 
    assumption that the image is composed of point sources.  This makes it a 
    poor choice for images with distributed flux.  In each iteration, a point 
    is added to the model at the location of the maximum residual, with a 
    fraction (specified by 'gain') of the magnitude.  The convolution of that 
    point is removed from the residual, and the process repeats.  Termination 
    happens after 'maxiter' iterations, or when the clean loops starts 
    increasing the magnitude of the residual.  This implementation can handle 
    1 and 2 dimensional data that is real valued or complex.
    gain: The fraction of a residual used in each iteration.  If this is too
        low, clean takes unnecessarily long.  If it is too high, clean does
        a poor job of deconvolving."""
    if mdl is None:
        mdl = np.zeros(im.shape, dtype=im.dtype)
        res = im.copy()
    else:
        mdl = mdl.copy()
        if len(mdl.shape) == 1:
            res = im - np.fft.ifft(np.fft.fft(mdl) * \
                                  np.fft.fft(ker)).astype(im.dtype)
        elif len(mdl.shape) == 2:
            res = im - np.fft.ifft2(np.fft.fft2(mdl) * \
                                   np.fft.fft2(ker)).astype(im.dtype)
        else: raise ValueError('Number of dimensions != 1 or 2')
    if area is None:
        area = np.ones(im.shape, dtype=np.int)
    else:
        area = area.astype(np.int)
        
    iter = _deconv.clean(res, ker, mdl, area,
            gain=gain, maxiter=maxiter, tol=tol, 
            stop_if_div=int(stop_if_div), verbose=int(verbose),
            pos_def=int(pos_def))
    score = np.sqrt(np.average(np.abs(res)**2))
    info = {'success':iter > 0 and iter < maxiter, 'tol':tol}
    if iter < 0: info.update({'term':'divergence', 'iter':-iter})
    elif iter < maxiter: info.update({'term':'tol', 'iter':iter})
    else: info.update({'term':'maxiter', 'iter':iter})
    info.update({'res':res, 'score':score})
    if verbose:
        print 'Term Condition:', info['term']
        print 'Iterations:', info['iter']
        print 'Score:', info['score']
    return mdl, info

def recenter(a, c):
    """Slide the (0,0) point of matrix a to a new location tuple c."""
    s = a.shape
    c = (c[0] % s[0], c[1] % s[1])
    a1 = np.concatenate([a[c[0]:], a[:c[0]]], axis=0)
    a2 = np.concatenate([a1[:,c[1]:], a1[:,:c[1]]], axis=1)
    return a2

def lsq(im, ker, mdl=None, area=None, gain=.1, tol=1e-3, maxiter=200, 
        lower=lo_clip_lev, upper=np.Inf, verbose=False):
    """This simple least-square fitting procedure for deconvolving an image 
    saves computing by assuming a diagonal pixel-pixel gradient of the fit.
    In essence, this assumes that the convolution kernel is a delta-function.
    This works for small kernels, but not so well for large ones.  See Cornwell 
    and Evans, 1984 "A Simple Maximum Entropy Deconvolution Algorithm" for more 
    information about this approximation.  Unlike maximum entropy, lsq makes 
    no promises about maximizing smoothness, but needs no information
    about noise levels.  Structure can be introduced for which there is no 
    evidence in the original image.  Termination happens when the fractional 
    score change is less than 'tol' between iterations.
    gain: The fraction of the step size (calculated from the gradient) taken
        in each iteration.  If this is too low, the fit takes unnecessarily 
        long.  If it is too high, the fit process can oscillate."""
    if mdl is None:
        #mdl = np.zeros_like(im)
        mdl = np.zeros(im.shape, dtype=im.dtype)
    x = mdl.copy()
    if area is None:
        area = np.ones(im.shape, dtype=np.int)
    else:
        area = area.astype(np.int)
    # Estimate gain of the kernel
    q = np.sqrt((np.abs(ker)**2).sum())
    ker_i = np.fft.fft2(ker)
    info = {'success':True, 'term':'maxiter', 'tol':tol}
    # Function to calculate chi-square and gradient
    def f(x):
        x_conv_ker = np.fft.ifft2(np.fft.fft2(x) * ker_i).real
        diff = (im - x_conv_ker) * area
        g_chi2 = -2*q*(diff)
        chi2 = np.abs(diff)**2
        return chi2, g_chi2
    score = 0
    # Start the fit loop
    for i in range(maxiter):
        chi2, g_chi2 = f(x)
        n_score = np.average(chi2)
        term = abs(1 - score/n_score)
        if verbose:
            slope = np.sqrt(np.average(g_chi2**2))
            print 'Step %d:' % i, 'score',  score, 
            print 'slope', slope, 'term', term
        if term < tol:
            info['term'] = 'tol'
            break
        score = n_score
        # For especially clean imgs, g_chi2 in some components can go to 0.
        # This check makes lsq a little slower for most images, though...
        d_x = np.where(abs(g_chi2) > 0, -(1/g_chi2) * chi2, 0)
        x = np.clip(x + gain * d_x, lower, upper)
    info.update({'res':im - np.fft.ifft2(np.fft.fft2(x) * ker_i).real,
        'score': score, 'iter':i+1})
    return x, info

def maxent(im, ker, var0, mdl=None, gain=.1, tol=1e-3, maxiter=200, 
        lower=lo_clip_lev, upper=np.Inf, verbose=False):
    """Maximum entropy deconvolution (MEM) (see Cornwell and Evans 1984
    "A Simple Maximum Entropy Deconvolution Algorithm" and Sault 1990
    "A Modification of the Cornwell and Evans Maximum Entropy Algorithm")
    is similar to lsq, but the fit is only optimized to within the specified
    variance (var0) and then "smoothness" is maximized.  This has several 
    desirable effects including uniqueness of solution, equal weighting of
    Fourier components, and absence of spurious structure.  The same 
    delta-kernel approximation (see lsq) is made here.
    var0: The estimated variance (noise power) in the image.  If none is
        provided, a quick lsq is used to estimate the variance of the residual.
    gain: The fraction of the step size (calculated from the gradient) taken
        in each iteration.  If this is too low, the fit takes unnecessarily 
        long.  If it is too high, the fit process can oscillate."""
    d_i = im.flatten()
    q = np.sqrt((ker**2).sum())
    minus_two_q = -2*q
    two_q_sq = 2*q**2
    if mdl is None:
         #mdl = np.ones_like(im) * np.average(im) / ker.sum() 
         mdl = np.ones(im.shape, dtype=im.dtype) * np.average(im) / ker.sum() 
    #if mdl is None: mdl = np.ones_like(im) * np.average(im) / q
    Nvar0 = d_i.size * var0
    inv_ker = np.fft.fft2(ker)
    m_i = mdl.flatten()
    def next_step(b_i, alpha, verbose=False):
        b_i.shape = im.shape
        b_i_conv_ker = np.fft.ifft2(np.fft.fft2(b_i) * inv_ker).real
        b_i.shape = m_i.shape
        diff = (im - b_i_conv_ker).flatten()
        chi2 = np.dot(diff,diff) - Nvar0
        g_chi2 = minus_two_q*(diff)
        g_J = (-np.log(b_i/m_i) - 1) - alpha * g_chi2
        gg_J = (-1/b_i) - alpha * two_q_sq
        # Define dot product using the metric of -gg_J^-1
        def dot(x, y): return (x*y/-gg_J).sum()
        score = dot(g_J, g_J) / dot(1,1)
        d_alpha = (chi2 + dot(g_chi2,g_J)) / dot(g_chi2,g_chi2)
        # For especially clean images, gg_J in some components can go to 0.
        # This check makes lsq a little slower for most images, though...
        d_b_i = np.where(abs(gg_J) > 0, -1/gg_J * (g_J - d_alpha*g_chi2), 0)
        if verbose:
            print '    score', score, 'fit', np.dot(diff,diff)
            print '    alpha', alpha, 'd_alpha', d_alpha
        return d_b_i, d_alpha, score
    alpha = 0.
    b_i = m_i.copy()
    info = {'success':True, 'term':'maxiter', 'var0':var0, 'tol':tol}
    for i in range(maxiter):
        if verbose: print 'Step %d:' % i
        d_b_i, d_alpha, score = next_step(b_i, alpha, verbose=verbose)
        if score < tol and score > 0:
            info['term'] = 'tol'
            break
        elif score > 1e10 or np.isnan(score) or score <= 0:
            info.update({'term':'divergence', 'success':False})
            break
        b_i = np.clip(b_i + gain * d_b_i, lower, upper)
        alpha += gain * d_alpha
    b_i.shape = im.shape
    info.update({'res':im - np.fft.ifft2(np.fft.fft2(b_i) * inv_ker).real,
        'score': score, 'alpha': alpha, 'iter':i+1})
    return b_i, info

def maxent_findvar(im, ker, var=None, f_var0=.6, mdl=None, gain=.1, tol=1e-3, 
        maxiter=200, lower=lo_clip_lev, upper=np.Inf, verbose=False, 
        maxiterok=False):
    """This frontend to maxent tries to find a variance for which maxent will
    converge.  If the starting variance (var) is not specified, it will be
    estimated as a fraction (f_var0) of the variance of the residual of a 
    lsq deconvolution, and then a search algorithm tests an ever-widening
    range around that value.  This function will search until it succeeds."""
    cl, info, cnt = None, None, -1
    if var is None:
        # Get a starting estimate of variance to use via residual of lsq
        junk, info = lsq(im, ker, mdl=mdl, gain=gain, tol=tol,
            maxiter=maxiter/4, lower=lower, upper=upper, verbose=False)
        var = np.var(info['res'])
        if verbose: print 'Using', f_var0, 'of LSQ estimate of var=', var
        var *= f_var0
    else:
        if verbose: print 'Using specified var=', var
    while cl is None:
        print cnt
        if cnt == -1: v = var
        else: v = var / (1.5**cnt)
        while cnt < 0 or v < var * (1.5**cnt):
            if verbose:
                print 'Trying var=', v,
                sys.stdout.flush()
            c, i = maxent(im, ker, v, mdl=mdl, gain=gain, tol=tol,
                maxiter=maxiter, lower=lower, upper=upper, verbose=False)
            if verbose:
                print 'success %d,' % i['success'],
                print 'term: %s,' % i['term'], 'score:' , i['score']
            # Check if fit converged
            if i['success'] and (maxiterok or i['term'] == 'tol'):
                cl, info = c, i
                break
            else:
                if not cl is None or cnt == -1: break
                v *= 1.2 ** (1./(2*(cnt+1)))
        cnt += 1
    if verbose: print 'Done with MEM.'
    return cl, info

def anneal(im, ker, mdl=None, maxiter=1000, lower=lo_clip_lev, upper=np.Inf,
        cooling=lambda i,x: 1e+1*(1-np.cos(i/50.))*(x**2), verbose=False):
    """Annealing takes a non-deterministic approach to deconvolution by
    randomly perturbing the model and selecting perturbations that improve the 
    residual.  By slowly reducing the temperature of the perturbations, 
    annealing attempts to settle into a global minimum.  Annealing is slower
    than lsq for a known gradient, but is less sensitive to gradient errors 
    (it can solve for wider kernels). Faster cooling speeds terminate more 
    quickly, but are less likely to find the global minimum.  This 
    implementation assigns a temperature to each pixel proportional to the 
    magnitude of the residual in that pixel and the global cooling speed.
    cooling: A function accepting (iteration,residuals) that returns a 
        vector of standard deviation for noise in the respective pixels.
        Picking the scaling of this function correctly is vital for annealing
        to work."""
    if mdl is None: mdl = np.zeros_like(im)
    q = np.sqrt((ker**2).sum())
    inv_ker = np.fft.fft2(ker)
    dif = im - np.fft.ifft2(np.fft.fft2(mdl) * inv_ker).real
    score = np.average(dif**2)
    #info = {'success':True, 'term':'maxiter', 'speed':speed}
    info = {'success':True, 'term':'maxiter'}
    for i in range(maxiter):
        delta = np.random.normal(scale=1., size=mdl.shape) * cooling(i, dif/q)
        n_mdl = np.clip(mdl + delta, lower, upper)
        n_dif = im - np.fft.ifft2(np.fft.fft2(n_mdl) * inv_ker).real
        n_score = np.average(n_dif**2)
        if verbose: print 'Step %d:' % i, n_score, score
        if n_score < score: mdl, dif, score = n_mdl, n_dif, n_score
    info.update({'res':dif, 'score': score, 'iter':i+1})
    return mdl, info



# -*- coding: utf-8 -*-
import unittest, aipy as a, numpy as n, re
from aipy.miriad import ij2bl

class TestParseAnts(unittest.TestCase):
    'Tests aipy.scripting.parse_ants()'
    def testant4(self):
        """Test aipy.scripting.parse_ants()"""
        nants = 4
        cases = {
            'all': [],
            'auto': [('auto',1)],
            'cross': [('auto',0)],
            '0_1': [(ij2bl(0,1),1)],
            '0_1,1_2': [(ij2bl(0,1),1), (ij2bl(1,2),1)],
            '0x_1x': [(ij2bl(0,1),1,'xx')],
            '(0x,0y)_1x': [(ij2bl(0,1),1,'xx'), (ij2bl(0,1),1,'yx')],
            '(0,1)_2': [(ij2bl(0,2),1), (ij2bl(1,2),1)],
            '0_(1,2)': [(ij2bl(0,1),1), (ij2bl(0,2),1)],
            '(0,1)_(2,3)': [(ij2bl(0,2),1), (ij2bl(0,3),1),
                            (ij2bl(1,2),1), (ij2bl(1,3),1)],
            '0_(1,-2)': [(ij2bl(0,1),1), (ij2bl(0,2),0)],
            '(-0,1)_(2,-3)': [(ij2bl(0,2),0), (ij2bl(0,3),0), (ij2bl(1,2),1), (ij2bl(1,3),0)],
            '0,1,all': [],
        }
        for i in range(nants):
            cases[str(i)] = map(lambda x: (ij2bl(x,i),1), range(nants))
            cases['-'+str(i)] = map(lambda x: (ij2bl(x,i),0), range(nants))
        # inelegantly paste on the new pol parsing flag on the above tests
        # XXX really should add some new tests for the new pol parsing
        for k in cases:
            cases[k] = [(v+(-1,))[:3] for v in cases[k]]
        for ant_str in cases:
            self.assertEqual(a.scripting.parse_ants(ant_str, nants), 
                cases[ant_str])
        self.assertRaises(ValueError, a.scripting.parse_ants, '(0_1)_2', nants)

class TestParseSrcs(unittest.TestCase):
    'Tests aipy.scripting.parse_srcs()'
    def testsrc(self):
        """Test aipy.scripting.parse_srcs()"""
        cat = 'misc,helm'
        src_opt = 'cyg,cas,12_40,12:00_40:00,12:00:00.0_40:00:00.0,bug_man'
        srclist,cutoff,catalogs = a.scripting.parse_srcs(src_opt, cat)
        self.assertEqual(srclist[0], 'cyg')
        self.assertEqual(srclist[1], 'cas')
        self.assertEqual(srclist[2].src_name, '12_40')
        self.assertEqual(srclist[3].src_name, '12:00_40:00')
        self.assertEqual(srclist[4].src_name, '12:00:00.0_40:00:00.0')
        self.assertEqual(srclist[5], 'bug_man')

class TestParseChans(unittest.TestCase):
    'Tests aipy.scripting.parse_chans()'
    def setUp(self):
        self.nchan = 256
        self.cases = {
            'all': [n.arange(self.nchan)],
            '0,1,2,3': [n.array([0]),n.array([1]),n.array([2]),n.array([3])],
            '0_20': [n.arange(21)],
            '0_20_4': [n.arange(0,21,4)],
            '5,7,40_50': [n.array([5]),n.array([7]),n.arange(40,51)],
        }
    def testchan_concat(self):
        """Test aipy.scripting.parse_chans() - concatenate"""
        for case in self.cases:
            chans = a.scripting.parse_chans(case, self.nchan, concat=True)
            self.assertTrue(n.all(chans == n.concatenate(self.cases[case])))
        self.assertRaises(AssertionError,
            a.scripting.parse_chans, '0_1_2_3', self.nchan)
    def testchan_noconcat(self):
        """Test aipy.scripting.parse_chans() - without concatenate"""
        for case in self.cases:
            chans = a.scripting.parse_chans(case, self.nchan, concat=False)
            for ch1, ch2 in zip(chans, self.cases[case]):
                self.assertTrue(n.all(ch1 == ch2))

class TestParsePrms(unittest.TestCase):
    def setUp(self):
        self.test_strs = [
            'src1=jys',
            'src2=jys/100',
            'src3=jys/100/1',
            'src4=jys//1',
            '(src5/src6)=jys',
            '(src7/src8)=jys/(100/200)',
            '(src9/src10)=jys/(100/200)/(1/2)',
            '(src11/src12)=jys/(100/200)/1',
            '(src13/src14/src15)=(jys/index)//1',
        ]
    def testname(self):
        """Test aipy.scripting.name matching"""
        r = re.compile(a.scripting.name)
        self.assertEqual(r.match('src1').groups(), ('src1',))
        self.assertEqual(r.match('1').groups(), ('1',))
        self.assertEqual(r.match('1,1').groups(), ('1',))
        self.assertEqual(r.match('1=1').groups(), ('1',))
        self.assertEqual(r.match('1/1').groups(), ('1',))
        self.assertEqual(r.match('1(1').groups(), ('1',))
        self.assertEqual(r.match('1)1').groups(), ('1',))
        self.assertEqual(r.match(')1'), None)
    def testgrp(self):
        """Test aipy.scripting.grp matching"""
        r = re.compile(a.scripting.grp)
        self.assertEqual(r.match('src1').groups()[0], 'src1')
        self.assertEqual(r.match('src1/src2').groups()[0], 'src1')
        self.assertEqual(r.match('(src1/src2)').groups()[0], '(src1/src2)')
        self.assertEqual(r.match('(1/2/3)').groups()[0], '(1/2/3)')
        self.assertEqual(r.match('(src1/)'), None)
        self.assertEqual(r.match('(src1/src2'), None)
    def testprm(self):
        """Test aipy.scripting.prm_rgx matching"""
        r = a.scripting.prm_rgx
        for t in self.test_strs:
            self.assertEqual(r.match(t).groups()[0], t)
    def testprmlist(self):
        """Test aipy.scripting.parse_prms()"""
        t = ','.join(self.test_strs)
        prms = a.scripting.parse_prms(t)
        self.assertEqual(prms['src1']['jys'], (None,None))
        self.assertEqual(prms['src2']['jys'], (100.,None))
        self.assertEqual(prms['src3']['jys'], (100.,1.))
        self.assertEqual(prms['src4']['jys'], (None,1.))
        self.assertEqual(prms['src5']['jys'], (None,None))
        self.assertEqual(prms['src6']['jys'], (None,None))
        self.assertEqual(prms['src7']['jys'], (100.,None))
        self.assertEqual(prms['src8']['jys'], (200.,None))
        self.assertEqual(prms['src9']['jys'], (100.,1.))
        self.assertEqual(prms['src10']['jys'], (200.,2.))
        self.assertEqual(prms['src11']['jys'], (100.,1.))
        self.assertEqual(prms['src12']['jys'], (200.,1.))
        self.assertEqual(prms['src13']['jys'], (None,1.))
        self.assertEqual(prms['src13']['index'], (None,1.))
        self.assertEqual(prms['src14']['jys'], (None,1.))
        self.assertEqual(prms['src14']['index'], (None,1.))
        self.assertEqual(prms['src15']['jys'], (None,1.))
        self.assertEqual(prms['src15']['index'], (None,1.))
        self.assertRaises(AssertionError, 
            a.scripting.parse_prms,'(a/b)=(c/d)/(1/2)/(3/4)')
        t = '(1/2/3)=jys,(2/3)=index'
        prms = a.scripting.parse_prms(t)
        self.assertEqual(len(prms['1']), 1)
        self.assertEqual(len(prms['2']), 2)
        self.assertEqual(len(prms['3']), 2)
        prms = a.scripting.parse_prms('a=(b/c)')
        self.assertEqual(len(prms['a']), 2)

class TestSuite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the aipy.coord unit tests."""

    def __init__(self):
        unittest.TestSuite.__init__(self)

        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(TestParseAnts))
        self.addTests(loader.loadTestsFromTestCase(TestParseSrcs))
        self.addTests(loader.loadTestsFromTestCase(TestParseChans))
        self.addTests(loader.loadTestsFromTestCase(TestParsePrms))

if __name__ == '__main__':
    unittest.main()




# -*- coding: utf-8 -*-
import unittest, aipy._alm as a, numpy as n

class TestAlm(unittest.TestCase):
    def test_init(self):
        """Test initializing a aipy._alm object"""
        alm = a.Alm(10, 5)
        self.assertEqual(alm.lmax(), 10)
        self.assertEqual(alm.mmax(), 5)
        self.assertRaises(RuntimeError, a.Alm, 5, 10)

if False:
  class TestMemLeaks(unittest.TestCase):
    def setUp(self):
        self.alm = a.Alm(10,10)
    def test_alm_create(self):
        while True: alm = a.Alm(20,20)
    def test_alm_to_map(self):
        while True: d = self.alm.to_map(256, 'RING')
    def test_alm_from_map(self):
        d = n.zeros(12*256**2, dtype=n.float)
        while True: self.alm.from_map(d, 2)
    def test_alm_get_data(self):
        while True: c = self.alm.get_data()
    def test_alm_set_data(self):
        c = self.alm.get_data()
        while True: self.alm.set_data(c)

class TestSuite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the aipy._alm unit tests."""

    def __init__(self):
        unittest.TestSuite.__init__(self)

        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(TestAlm))
        #self.addTests(loader.loadTestsFromTestCase(TestMemLeaks))

if __name__ == '__main__':
    unittest.main()



"""Load calibration information contained in specified calibration
modules."""

import numpy as np
import os
import sys
from . import src

def get_freqs(sdf, sfreq, nchan):
    return np.arange(nchan, dtype=np.float) * sdf + sfreq

def get_aa(*args):
    '''Return the AntennaArray specified by cal_key, which should be the
    name of a module somewhere in Python's path that implements a 
    "get_aa(freqs)" function.  That function should return an AntennaArray 
    initialized with appropriate calibration parameters.  This function simply 
    attempts to import that function, and then performs a simple conversion 
    between sdf (channel width in GHz), sfreq (starting frequency), nchan (#
    of channels) --> freqs (an array of channel centers).'''
    if len(args) == 2: cal_key,freqs = args
    else:
        cal_key,sdf,sfreq,nchan = args
        freqs = get_freqs(sdf, sfreq, nchan)
    sys.path.append(os.getcwd())
    exec('from %s import get_aa as _get_aa' % cal_key)
    return _get_aa(freqs)

def get_catalog(cal_key=None, srcs=None, cutoff=None, catalogs=['helm','misc']):
    '''Return the source catalog specified by cal_key, which should be the
    name of a module somewhere in Python's path that implements a
    "get_catalog(srcs, cutoff, catalogs)" function.  That function should return a
    SrcCatalog initialized with appropriate calibration parameters.  This
    function simply attempts to import that function and then pass it the
    specified srcs, cutoff.  If no such function is found, src.get_catalog()
    is called instead.'''
    sys.path.append(os.getcwd())
    try: exec('from %s import get_catalog as _get_catalog' % cal_key)
    except(ImportError): _get_catalog = src.get_catalog
    return _get_catalog(srcs=srcs, cutoff=cutoff, catalogs=catalogs)



"""
Module for gridding UVW data (including W projection), forming images,
and combining (mosaicing) images into spherical maps.
"""

import numpy as np, utils, coord, time
try:
    from astropy.io import fits as pyfits
except ImportError:
    import pyfits
USEDSP = True
if USEDSP: import _dsp

deg2rad = np.pi / 180.
rad2deg = 180. / np.pi

def word_wrap(string, width=80, ind1=0, ind2=0, prefix=''):
    """ 
    word wrapping function.
        string: the string to wrap
        width: the column number to wrap at
        prefix: prefix of each line with thie string (goes before any indentation)
        ind1: number of characters to indent the first line
        ind2: number of characters to indent the rest of the lines
    """
    awidth = min(width-2-len(prefix+ind1*' '),width-2-len(prefix+ind2*' '))
    words = string.split(' ')
    okwords = []
    chunk = lambda v,l: [v[i*l:(i+1)*l] for i in range(int(np.ceil(len(v)/float(l))))]
    for word in words:
        for okword in chunk(word,awidth):
            okwords.append(okword)
    lines = []
    l = prefix+ind1*' '
    for i,w in enumerate(okwords):
        if len(l+' ' + w)<width:
            l += ' '+w
        else:
            lines.append(l)
            l = prefix + ind2*' '+w
    lines.append(l)
    return '\n'.join(lines)

def recenter(a, c):
    """Slide the (0,0) point of matrix a to a new location tuple c.  This is
    useful for making an image centered on your screen after performing an
    inverse fft of uv data."""
    s = a.shape
    c = (c[0] % s[0], c[1] % s[1])
    if np.ma.isMA(a):
        a1 = np.ma.concatenate([a[c[0]:], a[:c[0]]], axis=0)
        a2 = np.ma.concatenate([a1[:,c[1]:], a1[:,:c[1]]], axis=1)
    else:
        a1 = np.concatenate([a[c[0]:], a[:c[0]]], axis=0)
        a2 = np.concatenate([a1[:,c[1]:], a1[:,:c[1]]], axis=1)
    return a2

def convolve2d(a, b):
    """Convolve a and b by multiplying in Fourier domain.  Must be same size."""
    return np.fft.ifft2(np.fft.fft2(a) * np.fft.fft2(b))

def gaussian_beam(sigma, shape=0, amp=1., center=(0,0)):
    """Return a 2D gaussian.  Normalized to area under curve = 'amp'.  
    Down by 1/e at distance 'sigma' from 'center'."""
    if type(shape) == type(0): shape = np.array([2, 2]) * sigma
    def gaussian(x, y):
        nx = np.where(x > shape[0] / 2, x - shape[0], x)
        ny = np.where(y > shape[1] / 2, y - shape[1], y)
        return np.exp(-(nx**2 + ny**2) / sigma**2)
    g = np.fromfunction(gaussian, shape)
    g *= amp
    return recenter(g, center)

def beam_gain(bm):
    return np.abs(bm).max()

class Img:
    """Class for gridding uv data, recording the synthesized beam profile,
    and performing transforms into image domain."""
    def __init__(self, size=100, res=1, mf_order=0):
        """size = number of wavelengths which the UV matrix spans (this 
        determines the image resolution).
        res = resolution of the UV matrix (determines image field of view)."""
        self.res = float(res)
        self.size = float(size)
        dim = np.round(self.size / self.res)
        self.shape = (int(dim),int(dim))
        self.uv = np.zeros(shape=self.shape, dtype=np.complex64)
        self.bm = []
        for i in range(mf_order+1):
            self.bm.append(np.zeros(shape=self.shape, dtype=np.complex64))
    def get_LM(self, center=(0,0)):
        """Get the (l,m) image coordinates for an inverted UV matrix."""
        dim = self.shape[0]
        M,L = np.indices(self.shape)
        L,M = np.where(L > dim/2, dim-L, -L), np.where(M > dim/2, M-dim, M)
        L,M = L.astype(np.float32)/dim/self.res, M.astype(np.float32)/dim/self.res
        mask = np.where(L**2 + M**2 >= 1, 1, 0)
        L,M = np.ma.array(L, mask=mask), np.ma.array(M, mask=mask)
        return recenter(L, center), recenter(M, center)
    def get_indices(self, u, v):
        """Get the pixel indices corresponding to the provided uv coordinates."""
        if not USEDSP:
            u = np.round(u / self.res).astype(np.int)
            v = np.round(v / self.res).astype(np.int)
            return np.array([-v,u],).transpose()
        else:
            return (-v / self.res).astype(np.float32), (u / self.res).astype(np.float32)
    def get_uv(self):
        """Return the u,v indices of the pixels in the uv matrix."""
        u,v = np.indices(self.shape)
        u = np.where(u < self.shape[0]/2, u, u - self.shape[0])
        v = np.where(v < self.shape[1]/2, v, v - self.shape[1])
        return u*self.res, v*self.res
    def put(self, (u,v,w), data, wgts=None, apply=True):
        """Grid uv data (w is ignored) onto a UV plane.  Data should already
        have the phase due to w removed.  Assumes the Hermitian conjugate
        data is in uvw already (i.e. the conjugate points are not placed for
        you).  If wgts are not supplied, default is 1 (normal weighting).
        If apply is false, returns uv and bm data without applying it do
        the internally stored matrices."""
        if wgts is None:
            wgts = []
            for i in range(len(self.bm)):
                if i == 0: wgts.append(np.ones_like(data))
                else: wgts.append(np.zeros_like(data))
        if len(self.bm) == 1 and len(wgts) != 1: wgts = [wgts]
        assert(len(wgts) == len(self.bm))
        if apply: uv,bm = self.uv,self.bm
        else:
            uv = np.zeros_like(self.uv)
            bm = [np.zeros_like(i) for i in self.bm]
        if not USEDSP:
            inds = self.get_indices(u,v)
            
            ok = np.logical_and(np.abs(inds[:,0]) < self.shape[0],
                np.abs(inds[:,1]) < self.shape[1])
            data = data.compress(ok)
            inds = inds.compress(ok, axis=0)
            utils.add2array(uv, inds, data.astype(uv.dtype))
        else:
            u,v = self.get_indices(u,v)
            _dsp.grid2D_c(uv, u, v, data.astype(uv.dtype))
        
        for i,wgt in enumerate(wgts):
            if not USEDSP:
                wgt = wgt.compress(ok)
                utils.add2array(bm[i], inds, wgt.astype(bm[0].dtype))
            else:
                _dsp.grid2D_c(bm[i], u, v, wgt.astype(bm[0].dtype))
        if not apply: return uv, bm
    def get(self, (u,v,w), uv=None, bm=None):
        """Generate data as would be observed at the provided (u,v,w) based on
        this Img's current uv data.  Phase due to 'w' will be applied to data
        before returning."""
        u,v = u.flatten(), v.flatten()
        if uv is None: uv,bm = self.uv, self.bm[0]
        if not USEDSP:
            # Currently: no interpolation
            inds = self.get_indices(u,v)
            data = self.uv[inds[:,0],inds[:,1]] / self.bm[0][inds[:,0],inds[:,1]]
            data = data.squeeze()
        else:
            u,v = self.get_indices(-u,v)
            u,v = -v,u # XXX necessary, but probably because of axis ordering in FITS files...
            uvdat = np.zeros(u.shape, dtype=np.complex64)
            bmdat = np.zeros(u.shape, dtype=np.complex64)
            _dsp.degrid2D_c(uv, u, v, uvdat)
            _dsp.degrid2D_c(bm, u, v, bmdat)
            #data = uvdat.sum() / bmdat.sum()
            data = uvdat / bmdat
        return data
    def append_hermitian(self, (u,v,w), data, wgts=None):
        """Append to (uvw, data, [wgts]) the points (-uvw, conj(data), [wgts]).
        This is standard practice to get a real-valued image."""
        u = np.concatenate([u, -u], axis=0)
        v = np.concatenate([v, -v], axis=0)
        w = np.concatenate([w, -w], axis=0)
        data = np.concatenate([data, np.conj(data)], axis=0)
        if wgts is None: return np.array((u,v,w)), data
        if len(self.bm) == 1 and len(wgts) != 1: wgts = [wgts]
        assert(len(wgts) == len(self.bm))
        for i,wgt in enumerate(wgts): wgts[i] = np.concatenate([wgt,wgt],axis=0)
        return (u,v,w), data, wgts
    def _gen_img(self, data, center=(0,0)):
        """Return the inverse FFT of the provided data, with the 0,0 point 
        moved to 'center'.  Up=North, Right=East."""
        return recenter(np.fft.ifft2(data).real.astype(np.float32), center)
    def image(self, center=(0,0)):
        """Return the inverse FFT of the UV matrix, with the 0,0 point moved
        to 'center'.  Tranposes to put up=North, right=East."""
        return self._gen_img(self.uv, center=center)
    def bm_image(self, center=(0,0), term=None):
        """Return the inverse FFT of the sample weightings (for all mf_order
        terms, or the specified term if supplied), with the 0,0 point
        moved to 'center'.  Tranposes to put up=North, right=East."""
        if not term is None:
            return self._gen_img(self.bm[term], center=center)
        else:
            return [self._gen_img(b, center=center) for b in self.bm]
    def get_top(self, center=(0,0)):
        """Return the topocentric coordinates of each pixel in the image."""
        x,y = self.get_LM(center)
        z = np.sqrt(1 - x**2 - y**2)
        return x,y,z
    def get_eq(self, ra=0, dec=0, center=(0,0)):
        """Return the equatorial coordinates of each pixel in the image, 
        assuming the image is centered on the provided ra, dec (in radians)."""
        x,y,z = self.get_top(center)
        shape,mask = x.shape, x.mask
        if len(mask.shape) == 0: mask = np.zeros(x.shape)
        vec = np.array([a.filled().flatten() for a in (x,y,z)])
        m = coord.top2eq_m(-ra, dec)
        vec = np.dot(m, vec)
        vec.shape = (3,) + shape
        return np.ma.array(vec, mask=[mask,mask,mask])

class ImgW(Img):
    """A subclass of Img adding W projection functionality (see Cornwell
    et al. 2005 "Widefield Imaging Problems in Radio Astronomy")."""
    def __init__(self, size=100, res=1, wres=.5, mf_order=0):
        """wres: the gridding resolution of sqrt(w) when projecting to w=0."""
        Img.__init__(self, size=size, res=res, mf_order=mf_order)
        self.wres = wres
        self.wcache = {}
    def put(self, (u,v,w), data, wgts=None, invker2=None):
        """Same as Img.put, only now the w component is projected to the w=0
        plane before applying the data to the UV matrix."""
        if len(u) == 0: return
        if wgts is None:
            wgts = []
            for i in range(len(self.bm)):
                if i == 0: wgts.append(np.ones_like(data))
                else: wgts.append(np.zeros_like(data))
        if len(self.bm) == 1 and len(wgts) != 1: wgts = [wgts]
        assert(len(wgts) == len(self.bm))
        # Sort uvw in order of w
        order = np.argsort(w)
        u = u.take(order)
        v = v.take(order)
        w = w.take(order)
        data = data.take(order)
        wgts = [wgt.take(order) for wgt in wgts]
        sqrt_w = np.sqrt(np.abs(w)) * np.sign(w)
        i = 0
        while True:
            # Grab a chunk of uvw's that grid w to same point.
            j = sqrt_w.searchsorted(sqrt_w[i]+self.wres)
            print '%d/%d datums' % (j, len(w))
            avg_w = np.average(w[i:j])
            # Put all uv's down on plane for this gridded w point
            wgtsij = [wgt[i:j] for wgt in wgts]
            uv,bm = Img.put(self, (u[i:j],v[i:j],w[i:j]),
                data[i:j], wgtsij, apply=False)
            # Convolve with the W projection kernel
            invker = np.fromfunction(lambda u,v: self.conv_invker(u,v,avg_w),
                uv.shape)
            if not invker2 is None: invker *= invker2
            self.uv += np.fft.ifft2(np.fft.fft2(uv) * invker)
            for b in range(len(self.bm)):
                self.bm[b] += np.fft.ifft2(np.fft.fft2(bm[b]) * invker)
            if j >= len(w): break
            i = j
    def get(self, (u,v,w)):
        order = np.argsort(w.flat)
        u_,v_,w_ = u.take(order).squeeze(), v.take(order).squeeze(), w.take(order).squeeze()
        sqrt_w = np.sqrt(np.abs(w_)) * np.sign(w_)
        i, d_ = 0, []
        while True:
            # Grab a chunk of uvw's that grid w to same point.
            j = sqrt_w.searchsorted(sqrt_w[i]+self.wres)
            #print j, len(sqrt_w)
            id = np.round(np.average(sqrt_w[i:j]) / self.wres) * self.wres
            if not self.wcache.has_key(id):
                avg_w = np.average(w_[i:j])
                print 'Caching W plane ID=', id
                projker = np.fromfunction(lambda us,vs: self.conv_invker(us,vs,-avg_w), 
                    self.uv.shape).astype(np.complex64)
                uv_wproj = np.fft.ifft2(np.fft.fft2(self.uv) * projker).astype(np.complex64)
                bm_wproj = np.fft.ifft2(np.fft.fft2(self.bm[0]) * projker).astype(np.complex64) # is this right to convolve?
                self.wcache[id] = (uv_wproj, bm_wproj)
                print '%d W planes cached' % (len(self.wcache))
            # Put all uv's down on plane for this gridded w point
            uv_wproj, bm_wproj = self.wcache[id]
            # Could think about improving this by interpolating between w planes.
            d_.append(Img.get(self, (u_[i:j],v_[i:j],w_[i:j]), uv_wproj, bm_wproj))
            if j >= len(sqrt_w): break
            i = j
        d_ = np.concatenate(d_)
        # Put back into original order
        deorder = np.argsort(order)
        return d_.take(deorder)
    def conv_invker(self, u, v, w):
        """Generates the W projection kernel (a function of u,v) for the
        supplied value of w.  See Cornwell et al. 2005 "Widefield Imaging
        Problems in Radio Astronomy" for discussion.  This implementation
        uses a numerically evaluated Fresnel kernel, rather than the
        small-angle approximated one given in the literature."""
        L,M = self.get_LM()
        # This is the exactly evaluated kernel (works better)
        sqrt = np.sqrt(1 - L**2 - M**2).astype(np.complex64)
        G = np.exp(-2*np.pi*1j*w*(sqrt - 1))
        # This is the kernel described by Cornwell using the small angle approx.
        #G = np.exp(np.pi*1j*w*(l**2 + m**2))
        G = G.filled(0)
        # Unscramble difference between fft(fft(G)) and G
        G[1:] = np.flipud(G[1:]).copy()
        G[:,1:] = np.fliplr(G[:,1:]).copy()
        return G / G.size

default_fits_format_codes = {
    np.bool_:'L', np.uint8:'B', np.int16:'I', np.int32:'J', np.int64:'K',
    np.float32:'E', np.float64:'D', np.complex64:'C', np.complex128:'M'
}

def to_fits(filename, data, clobber=False,
        axes=('ra--sin','dec--sin'),
        object='', telescope='', instrument='', observer='', origin='AIPY',
        obs_date=time.strftime('%D'), cur_date=time.strftime('%D'), 
        ra=0, dec=0, d_ra=0, d_dec=0, epoch=2000., 
        freq=0, d_freq=0, bscale=0, bzero=0,history=''):
    """Write image data to a FITS file.  Follows convention of VLA image
    headers.  "axes" describes dimensions of "data" provided.  (ra,dec) are
    the degree coordinates of image center in the specified "epoch". 
    (d_ra,d_dec) are approximate pixel-deltas for ra,dec (approximate because 
    if sine projections of these coordinates are specified--e.g. 
    "ra---sin"--then the deltas change away from the image center).  If a 
    "freq" axis is specified, then "freq" is the frequency of the first entry 
    (in Hz), and "d_freq" is the width of the channel.  The rest are pretty 
    self-explanitory/can be used however you want."""
    data = data.squeeze()
#    data.shape = (1,) * (len(axes) - len(data.shape)) + data.shape
    if len(data.shape) != len(axes):
        raise TypeError('to_fits: axes dimension list does not match data shape')
    phdu = pyfits.PrimaryHDU(data)
    phdu.update_header()
    phdu.header.update('TRANSPOS',0,comment='Import code for old AIPY convention.')
    phdu.header.update('OBJECT', object, comment='SOURCE NAME')
    phdu.header.update('TELESCOP', telescope)
    phdu.header.update('INSTRUME', instrument)
    phdu.header.update('OBSERVER', observer)
    phdu.header.update('DATE-OBS', obs_date, 
        comment='OBSERVATION START DATE DD/MM/YY')
    phdu.header.update('BSCALE ', bscale,
        comment='REAL = FITS_VALUE * BSCALE + BZERO')
    phdu.header.update('BZERO  ', bzero)
    phdu.header.update('BUNIT  ', 'JY/BEAM ', comment='UNITS OF FLUX')
    phdu.header.update('EQUINOX', epoch, comment='EQUINOX OF RA DEC')
    phdu.header.update('EPOCH',epoch,'Epoch of coordinate system')
    phdu.header.update('DATAMAX', data.max(), comment='MAX PIXEL VALUE')
    phdu.header.update('DATAMIN', data.min(), comment='MIN PIXEL VALUE')
    for i,ax in enumerate(axes):
        if ax.lower().startswith('ra') or ax.lower().startswith('glon'): val,delta = (ra, d_ra)
        elif ax.lower().startswith('dec') or ax.lower().startswith('glat'): val,delta = (dec, d_dec)
        elif ax.lower().startswith('freq'): val,delta = (freq, d_freq)
        elif ax.lower().startswith('stokes'): val,delta = (1, 1)
        else: val,delta = (0,0)
        phdu.header.update('CTYPE%d' % (i+1), ax.upper())
        if ax.lower().startswith('ra') or ax.lower().startswith('dec')\
        or ax.lower().startswith('glon') or ax.lower().startswith('glat'):
            phdu.header.update('CRPIX%d' % (i+1), np.ceil(phdu.data.shape[-(i+1)]/2.))
        else:
            phdu.header.update('CRPIX%d' % (i+1), np.ceil(phdu.data.shape[-(i+1)]/2.))
        phdu.header.update('CRVAL%d' % (i+1), val)
        if (ax.lower().startswith('ra') or ax.lower().startswith('glon')) and delta>0:
            phdu.header.update('CDELT%d' % (i+1), delta*-1)
        else:
            phdu.header.update('CDELT%d' % (i+1), delta)
        phdu.header.update('CROTA%d' % (i+1), 0)
        phdu.header.update('NAXIS%d' % (i+1),phdu.data.shape[i])
    if history!='':
        history = [h.strip() for h in history.split("\n")]
        for line in history:
            if len(line)>1:
                if line.startswith('#'):
                    for subline in word_wrap(line,80,1,1,'').split("\n"):
                        phdu.header.add_history(subline)
                else:
                    for subline in word_wrap(line,70,5,10,'#').split("\n"):
                        phdu.header.add_history(subline)
    phdu.header.update('ORIGIN', origin)
    phdu.header.update('DATE', cur_date, comment='FILE WRITTEN ON DD/MM/YY')
    pyfits.writeto(filename,phdu.data,phdu.header,clobber=True)

def from_fits(filename):
    """Read (data,kwds) from a FITS file.  Matches to_fits() above.  Attempts
    to deduce each keyword listed in to_fits() from the FITS header, but is
    accepting of differences.  Returns values in "kwds" dictionary."""
    phdu = pyfits.open(filename)[0]
   
    try:
        if phdu.header['TRANSPOS']:
            data = phdu.data.transpose()
        else:
            data = phdu.data
    except(KeyError):
        data = phdu.data

    kwds = {}
    hitems = (('OBJECT','object'), ('TELESCOP','telescope'),
        ('INSTRUME','instrument'), ('OBSERVER','observer'),
        ('DATE-OBS','obs_date'), ('BSCALE','bscale'), ('BZERO','bzero'),
        ('EQUINOX','epoch'), ('ORIGIN','origin'), ('DATE','cur_date'))
    for fitsname,name in hitems:
        try: kwds[name] = phdu.header[fitsname]
        except(KeyError): pass
    axes = []
    for i in range(phdu.header['NAXIS']):
        try:
            ax = phdu.header['CTYPE%d' % (i+1)].lower()
            axes.append(ax)
            val = phdu.header['CRVAL%d' % (i+1)]
            delta = phdu.header['CDELT%d' % (i+1)]
            if ax.startswith('ra'): kwds['ra'],kwds['d_ra'] = (val,delta)
            elif ax.startswith('dec'): kwds['dec'],kwds['d_dec'] = (val,delta)
            elif ax.startswith('freq'): kwds['freq'],kwds['d_freq']=(val,delta)
            elif ax.startswith('stokes'): pass
            else: pass
        except(KeyError): pass
    kwds['axes'] = axes
    return data, kwds

def find_axis(phdu,name):
    """Find the axis number for RA"""
    for k in phdu.header.items():
        if k[0].lower().startswith('ctype'):
            if k[1].lower().startswith(name):
                return int(k[1][5])

def from_fits_to_fits(infile,outfile,data,kwds,history=None):
    """
    Create a fits file in outfile with data using header from infile using
    kwds to override values in the header.
    See img.to_fits for typical header variables.
    """

    phdu = pyfits.open(infile)[0]
    axes = []
    for i in range(1,phdu.header.get('naxis')+1):
        type = "CTYPE"+str(i)
        axes.append(phdu.header.get(type))
    print axes
    data.shape = data.shape #+ (1,) * (len(axes) - len(data.shape))
#    try: 
#        if phdu.header['TRANSPOS']:
#            data = phdu.data.transpose() #for backwards compatibility
#            phdu.header.update('TRANSPOS',0)
#        else:
#            data = phdu.data
#    except(KeyError):
#        data = phdu.data
    phdu.update_header()
    for i,ax in enumerate(axes):
        if ax.lower().startswith('ra'):
             if kwds.has_key('ra'): val=kwds['ra']
             else: val=None
             if kwds.has_key('d_ra'):delta = kwds['d_ra']
             else: delta=None
        elif ax.lower().startswith('dec'):
             if kwds.has_key('dec'): val=kwds['dec']
             else: val=None
             if kwds.has_key('d_dec'):delta = kwds['d_dec']
             else: delta=None
        elif ax.lower().startswith('freq'):
             if kwds.has_key('freq'): val=kwds['freq']
             else: val=None
             if kwds.has_key('d_freq'):delta = kwds['d_freq']
             else: delta=None
        else: val,delta = None,None
        phdu.header.update('CTYPE%d' % (i+1), ax.upper())
        if ax.lower().startswith('ra') or ax.lower().startswith('dec'):
            phdu.header.update('CRPIX%d' % (i+1), 
                    round(data.shape[-(i+1)]/2.))
        else:
            phdu.header.update('CRPIX%d' % (i+1), 1)
        print ax,round(data.shape[-(i+1)]/2.)
        if not val is None: phdu.header.update('CRVAL%d' % (i+1), val);
        if not delta is None: phdu.header.update('CDELT%d' % (i+1), delta)
        phdu.header.update('CROTA%d' % (i+1), 0)
    for k,v in kwds.iteritems():
        try:
            phdu.header.update(k,v)
        except(ValueError): 
            continue
    if history is None:history = "from_fits_to_fits: from %s to %s"%(infile,
                outfile)
    history = history.split("\n")
    for line in history:
        if len(line)>1:
            for subline in word_wrap(line,70,5,10,'#').split("\n"):
                phdu.header.add_history(subline)
    pyfits.writeto(outfile, data, phdu.header, clobber=True)



#! /usr/bin/env python
"""
Creates waterfall plots from Miriad UV files.  Can tile multiple plots
on one window, or plot just a single baseline.  When taking the delay
transform (-d), channels are interepreted as selecting delays after the
the transform operation.  Similarly, the time select (-t) will be interpreted
as selecting fringe rates if the fringe-rate transform (-f) has been selected.
In both cases, the ranges specified are intepreted to be the units of the
output plot (i.e. as specified by --chan_axis and --time_axis).

Author: Aaron Parsons, Griffin Foster
"""

import aipy as a, numpy as np, sys, optparse
from matplotlib import pylab as p

o = optparse.OptionParser()
o.set_usage('plot_uv.py [options] *.uv')
o.set_description(__doc__)
a.scripting.add_standard_options(o, src=True,ant=True, pol=True, chan=True, dec=True,
    cmap=True, max=True, drng=True, cal=True)
o.add_option('-m', '--mode', dest='mode', default='log',
    help='Plot mode can be log (logrithmic), lin (linear), phs (phase), real, or imag.')
o.add_option('-t', '--time', dest='time', default='all', help='Select which time sample to plot. Options are: "all" (default), "<time1 #>_<time2 #>" (a range of times to plot), or "<time1 #>,<time2 #>" (a list of times to plot). If "all" or a range are selected, a 2-d image will be plotted. If a list of times is selected an xy plot will be generated.')
o.add_option('-u', '--unmask', dest='unmask', action='store_true',
    help='Plot masked data, too.')
o.add_option('-d', '--delay', dest='delay', action='store_true',
    help='Take FFT of frequency axis to go to delay (t) space.')
o.add_option('-f', '--fringe', dest='fringe', action='store_true',
    help='Take FFT of time axis to go to fringe (Hz) space.')
o.add_option('--dt', dest='dt', action='store_true',
    help='Remove a linear extrapolation from adjacent times.')
o.add_option('--df', dest='df', action='store_true',
    help='Remove a linear extrapolation from adjacent frequency channels.')
o.add_option('-o', '--out_file', dest='out_file', default='',
    help='If provided, will save the figure to the specified file instead of popping up a window.')
o.add_option('--time_axis', dest='time_axis', default='index',
    help='Choose time axis to be integration/fringe index (index), or physical coordinates (physical), or if doing xy plot in time-mode, (lst) is also available.  Default is index.')
o.add_option('--chan_axis', dest='chan_axis', default='index',
    help='Choose channel axis to be channel/delay index (index), or physical coordinates (physical).  Default is index.')
o.add_option('--clean', dest='clean', type='float',
    help='Deconvolve delay-domain data by the "beam response" that results from flagged data.  Specify a tolerance for termination (usually 1e-2 or 1e-3).')
o.add_option('--nolegend', dest='nolegend', action='store_true',
    help='Omit legend in last plot.')
o.add_option('--share', dest='share', action='store_true',
    help='Share plots in a single frame.')
o.add_option('--xlim', dest='xlim',
    help='Limits on the x axis (channel/delay) for plotting.')
o.add_option('--ylim', dest='ylim',
    help='Limits on the x axis (time/delay-rate) for plotting.')
o.add_option('--plot_each', dest='plot_each',
    help='Instead of a waterfall plot, plot each of the specified axis (chan,time)')
o.add_option('--window', dest='window', default='blackman-harris',
    help='Windowing function to use in delay transform.  Default is blackman-harris.  Options are: ' + ', '.join(a.dsp.WINDOW_FUNC.keys()))

def convert_arg_range(arg):
    """Split apart command-line lists/ranges into a list of numbers."""
    arg = arg.split(',')
    return [map(float, option.split('_')) for option in arg]

def gen_times(timeopt, uv, coords, decimate):
    if timeopt == 'all':
        def time_selector(t, cnt): return True
    else:
        timeopt = convert_arg_range(timeopt)
        if len(timeopt[0]) != 1:
            def time_selector(t, cnt):
                if coords == 'index': t = cnt
                for opt in timeopt:
                    if (t >= opt[0]) and (t < opt[1]): return True
                return False
        else:
            timeopt = [opt[0] for opt in timeopt]
            inttime = uv['inttime'] / a.const.s_per_day * decimate
            def time_selector(t, cnt):
                if coords == 'index': return cnt in timeopt
                for opt in timeopt:
                    if (t >= opt) and (t < opt + inttime): return True
                return False
    return time_selector

def data_mode(data, mode='abs'):
    if mode.startswith('phs'): data = np.angle(data.filled(0))
    elif mode.startswith('lin'):
        data = np.ma.absolute(data.filled(0))
        data = np.ma.masked_less_equal(data, 0)
    elif mode.startswith('real'): data = data.real
    elif mode.startswith('imag'): data = data.imag
    elif mode.startswith('log'):
        data = np.ma.absolute(data.filled(0))
        data = np.ma.masked_less_equal(data, 0)
        data = np.ma.log10(data)
    else: raise ValueError('Unrecognized plot mode.')
    return data

opts, args = o.parse_args(sys.argv[1:])

# Parse command-line options
cmap = p.get_cmap(opts.cmap)
if not opts.xlim == None: opts.xlim = map(float, opts.xlim.split('_'))
if not opts.ylim == None: opts.ylim = map(float, opts.ylim.split('_'))
uv = a.miriad.UV(args[0])
a.scripting.uv_selector(uv, opts.ant, opts.pol)
chans = a.scripting.parse_chans(opts.chan, uv['nchan'])
is_chan_range, is_time_range = True, True
if opts.plot_each == 'chan': is_chan_range = False
elif opts.plot_each == 'time': is_time_range = False
freqs = a.cal.get_freqs(uv['sdf'], uv['sfreq'], uv['nchan'])
freqs = freqs.take(chans)
if opts.delay:
    if freqs.size == freqs[-1] - freqs[0] + 1:
        # XXX someday could allow for equal spaced chans
        raise ValueError('Channels must be contiguous to do delay transform (chan=%s)' % (opts.chan))
    delays = np.fft.fftfreq(freqs.size, freqs[1]-freqs[0])
    delays = np.fft.fftshift(delays)
time_sel = gen_times(opts.time, uv, opts.time_axis, opts.decimate)
inttime = uv['inttime'] * opts.decimate
if not opts.src is None:
    srclist,cutoff,catalogs = a.scripting.parse_srcs(opts.src, opts.cat)
    src = a.cal.get_catalog(opts.cal, srclist, cutoff, catalogs).values()[0]
del(uv)

# Loop through UV files collecting relevant data
plot_x = {}
plot_t = {'jd':[], 'lst':[], 'cnt':[]}
times = []

# Hold plotting handles
plots = {}
plt_data = {}

for uvfile in args:
    print 'Reading', uvfile
    uv = a.miriad.UV(uvfile)
    if not opts.cal is None:
        aa = a.cal.get_aa(opts.cal, uv['sdf'], uv['sfreq'], uv['nchan'])
        aa.set_active_pol(opts.pol)
        aa.select_chans(chans)
    else: aa = None
    # Only select data that is needed to plot
    a.scripting.uv_selector(uv, opts.ant, opts.pol)
    uv.select('decimate', opts.decimate, opts.decphs)
    # Read data from a single UV file
    for (uvw,t,(i,j)),d in uv.all():
        bl = '%d,%d,%d' % (i,j,uv['pol'])
        if len(times) == 0 or times[-1] != t:
            times.append(t)
            # Implement time selection
            use_this_time = time_sel(t, (len(times)-1) / opts.decimate)
            if use_this_time:
                if aa == None: lst = uv['lst']
                else:
                    aa.set_jultime(t)
                    lst = aa.sidereal_time()
                plot_t['lst'].append(lst)
                plot_t['jd'].append(t)
                plot_t['cnt'].append(len(times)-1)
        if not use_this_time: continue
        d = d.take(chans)
        if opts.ant.find('%d_%d' % (j,i)) != -1: d = d.conj() # obey antenna ordering, if possible
        #apply cal phases
        if not opts.cal is None:
            aa.set_jultime(t)
            if not opts.src is None:
                src.compute(aa)
                d = aa.phs2src(d, src, i, j)
            #else: took out this mode because it's not used, and prefer not to phase.
            #    d *= np.exp(-1j*np.pi*aa.get_phs_offset(i,j))
        # Do delay transform if required
        if opts.delay:
            w = a.dsp.gen_window(d.shape[-1], window=opts.window)
            if opts.unmask:
                flags = np.ones(d.shape, dtype=np.float)
                d = d.data
            else:
                flags = np.logical_not(d.mask).astype(np.float)
                d = d.filled(0)
            d = np.fft.ifft(d*w)
            ker = np.fft.ifft(flags*w)
            gain = a.img.beam_gain(ker)
            if not opts.clean is None and not np.all(d == 0):
                d, info = a.deconv.clean(d, ker, tol=opts.clean)
                d += info['res'] / gain
            d = np.ma.array(d)
            d = np.fft.fftshift(d, axes=0)
        elif opts.unmask: d = d.data
        d.shape = (1,) + d.shape
        if not plot_x.has_key(bl): plot_x[bl] = []
        plot_x[bl].append(d)
    del(uv)

bls = plot_x.keys()
def sort_func(a, b):
    ai,aj,pa = map(int, a.split(','))
    bi,bj,pb = map(int, b.split(','))
    if bi > ai or (bi == ai and bj > aj) or (bi == ai and bj == aj and pb < pa): return -1
    return 1
bls.sort(cmp=sort_func)
if len(bls) == 0:
    print 'No data to plot.'
    sys.exit(0)
m2 = int(np.sqrt(len(bls)))
m1 = int(np.ceil(float(len(bls)) / m2))

# Generate all the plots
dmin,dmax = None, None
fig = p.figure()
if not opts.src is None:fig.suptitle(opts.src)
for cnt, bl in enumerate(bls):
    d = np.ma.concatenate(plot_x[bl], axis=0)
    i,j,pol = map(int,bl.split(','))
    if opts.df: d = d[:,:-2]/2 + d[:,2:]/2 - d[:,1:-1]
    if opts.dt: d = d[:-2]/2 + d[2:]/2 - d[1:-1]
    if opts.fringe:
        d = d.filled(0)
        w = a.dsp.gen_window(d.shape[0], window=opts.window); w.shape += (1,)
        wgts = np.where(d != 0, 1., 0.) * w
        gain = np.sqrt(np.average(wgts**2, axis=0))
        ker = np.fft.ifft(wgts, axis=0) # w already put in 2 lines above
        d = np.fft.ifft(d*w, axis=0)
        if not opts.clean is None:
            for chan in range(d.shape[1]):
                if gain[chan] == 0: continue
                d[:,chan],info = a.deconv.clean(d[:,chan],ker[:,chan],tol=opts.clean)
                d[:,chan] += info['res'] / gain[chan]
        d = np.fft.fftshift(d, axes=0)
        d = np.ma.array(d)
    plt_data[cnt+1] = d
    d = data_mode(d, opts.mode)
    if not opts.share:
        p.subplot(m2, m1, cnt+1)
        dmin,dmax = None,None
        label = ''
    else:
        pol = a.miriad.pol2str[pol]
        label = '%d%s,%d%s ' % (i,pol[0],j,pol[-1]) 
    if is_chan_range and is_time_range:
        if opts.fringe:
            if opts.time_axis == 'index':
                drates = np.fft.fftfreq(len(plot_t['cnt']), 1./len(plot_t['cnt']))
                step = drates[1] - drates[0]
                ylabel = 'Delay-Rate (bins)'
            else:
                drates = np.fft.fftfreq(len(plot_t['cnt']), inttime) * 1e3 # mHz
                step = drates[1] - drates[0]
                ylabel = 'Delay-Rate (milliHz)'
            drates = np.fft.fftshift(drates)
            t1,t2 = drates[0]-0.5*step,drates[-1]+0.5*step
        else:
            if opts.time_axis == 'index':
                step = 1.
                t1,t2 = plot_t['cnt'][0]-0.5*step, plot_t['cnt'][-1]+0.5*step
                ylabel = 'Time (integrations)'
            elif opts.time_axis=='lst':
                step = plot_t['lst'][1] - plot_t['lst'][0]
                t1,t2 = (plot_t['lst'][0]-0.5*step)*12/np.pi, (plot_t['lst'][-1]+0.5*step)*12/np.pi
                ylabel = 'Local Sideral time (hrs)'
            else:
                step = plot_t['jd'][1] - plot_t['jd'][0]
                t1,t2 = plot_t['jd'][0]-0.5*step, plot_t['jd'][-1]+0.5*step
                ylabel = 'Time (Julian Date)'
        if opts.delay:
            if opts.chan_axis == 'index':
                step = 1
                c1,c2 = len(chans)/2 - len(chans)-0.5*step, len(chans)/2-0.5*step
                xlabel = 'Delay (bins)'
            else:
                step = delays[1] - delays[0]
                c1,c2 = delays[0]-0.5*step, delays[-1]+0.5*step
                xlabel = 'Delay (ns)'
        else:
            if opts.chan_axis == 'index':
                step = 1
                c1,c2 = 0-0.5*step, len(chans)-0.5*step
                xlabel = 'Frequency (chan)'
            else:
                step = freqs[1] - freqs[0]
                c1,c2 = freqs[0]-0.5*step, freqs[-1]+0.5*step
                xlabel = 'Frequency (GHz)'
        if not opts.max is None: dmax = opts.max
        elif dmax is None: dmax = d.max()
        else: dmax = max(dmax,d.max())
        if not opts.drng is None: dmin = dmax - opts.drng
        elif dmin is None: dmin = d.min()
        else: dmin = min(dmin,d.min())
        plots[cnt+1] = p.imshow(d, extent=(c1,c2,t2,t1), origin='upper',
            aspect='auto', interpolation='nearest', 
            vmax=dmax, vmin=dmin, cmap=cmap)
        p.colorbar(shrink=0.5)
        p.xlabel(xlabel); p.ylabel(ylabel)
        if not opts.xlim == None: p.xlim(*opts.xlim)
        if not opts.ylim == None: p.ylim(opts.ylim[1],opts.ylim[0]) # Reverse b/c otherwise ylim flips origin for unknown reasons
    elif is_chan_range and not is_time_range:
        if opts.delay:
            if opts.chan_axis == 'index':
                plot_chans = range(len(chans)/2 - len(chans), len(chans)/2)
                xlabel = 'Delay (bins)'
            else:
                plot_chans = delays
                xlabel = 'Delay (ns)'
        else:
            if opts.chan_axis == 'index':
                plot_chans = chans
                xlabel = 'Frequency (chan)'
            else:
                plot_chans = freqs
                xlabel = 'Frequency (GHz)'
        if opts.time_axis == 'index':
            if cnt == 0: plot_t = plot_t['cnt']
            label += '#%d'
        else:
            if cnt == 0: plot_t = plot_t['jd']
            label += 'jd%f'
        for ti,t in enumerate(plot_t):
            p.plot(plot_chans, d[ti,:], '-', label=label % t)
        p.xlabel(xlabel)
        if not opts.xlim == None: p.xlim(*opts.xlim)
        if not opts.max is None: dmax = opts.max
        elif dmax is None: dmax = d.max()
        else: dmax = max(dmax,d.max())
        if not opts.drng is None: dmin = dmax - opts.drng
        elif dmin is None: dmin = d.min()
        else: dmin = min(dmin,d.min())
        if not opts.ylim == None: p.ylim(*opts.ylim)
        else: p.ylim(dmin,dmax)
    elif not is_chan_range and is_time_range:
        if opts.fringe:
            if opts.time_axis == 'index':
                drates = np.fft.fftfreq(len(plot_t['cnt']), 1./len(plot_t['cnt']))
                xlabel = 'Delay-Rate (bins)'
            else:
                print inttime, len(plot_t['cnt'])
                drates = np.fft.fftfreq(len(plot_t['cnt']), inttime) * 1e3 # mHz
                xlabel = 'Delay-Rate (milliHz)'
            plot_times = np.fft.fftshift(drates)
        else:
            if opts.time_axis == 'index':
                plot_times = range(len(plot_t['jd']))
                xlabel = 'Time (integrations)'
            elif opts.time_axis == 'physical':
                plot_times = plot_t['jd']
                xlabel = 'Time (JD)'
            elif opts.time_axis == 'lst':
                plot_times = plot_t['lst']
                xlabel = 'Time (Sidereal Radians)'
            else: raise ValueError('Unrecognized time axis type.')
        if opts.chan_axis == 'index': label += '#%d'
        else:
            chans = freqs
            label += '%f GHz'
        for c, chan in enumerate(chans):
            p.plot(plot_times, d[:,c], '-', label=label % chan)
        if not opts.max is None: dmax = opts.max
        elif dmax is None: dmax = d.max()
        else: dmax = max(dmax,d.max())
        if not opts.drng is None: dmin = dmax - opts.drng
        elif dmin is None: dmin = d.min()
        else: dmin = min(dmin,d.min())
        p.xlabel(xlabel)
        if not opts.xlim == None: p.xlim(*opts.xlim)
        if not opts.ylim == None: p.ylim(*opts.ylim)
        else: p.ylim(dmin,dmax)
    else: raise ValueError('Either time or chan needs to be a range.')
    if not opts.share:
        pol = a.miriad.pol2str[pol]
        title = '%d%s,%d%s ' % (i,pol[0],j,pol[-1]) 
        p.title(title)
if not opts.nolegend and (not is_time_range or not is_chan_range): 
    p.legend(loc='best')

# Save to a file or pop up a window
if opts.out_file != '': p.savefig(opts.out_file)
else:
    def click(event):
        print [event.key]
        if event.key == 'm':
            mode = raw_input('Enter new mode: ')
            for k in plots:
                try:
                    d = data_mode(plt_data[k], mode)
                    plots[k].set_data(d)
                except(ValueError):
                    print 'Unrecognized plot mode'
            p.draw()
        elif event.key == 'd':
            max = raw_input('Enter new max: ')
            try: max = float(max)
            except(ValueError): max = None
            drng = raw_input('Enter new drng: ')
            try: drng = float(drng)
            except(ValueError): drng = None
            for k in plots:
                _max,_drng = max, drng
                if _max is None or _drng is None:
                    d = plots[k].get_array()
                    if _max is None: _max = d.max()
                    if _drng is None: _drng = _max - d.min()
                plots[k].set_clim(vmin=_max-_drng, vmax=_max)
            print 'Replotting...'
            p.draw()
    p.connect('key_press_event', click)
    p.show()



#! /usr/bin/env python
import aipy as a, ephem as e, numpy as np, sys, optparse, os

o = optparse.OptionParser()
o.set_usage('modmap.py [options]')
o.set_description(__doc__)
a.scripting.add_standard_options(o, src=True, cal=True)
o.add_option('-m', '--map', dest='map', default='out.fits',
    help='The location to save the map.')
o.add_option('-i', '--in_map', dest='in_map', 
    help='Existing map to add sources to.  Treated as temperature, i.e. if resolution is degraded, pixels are averaged.  Use mscale to re-scale if you want Jy per pixel.')
o.add_option('-f', '--freq', dest='freq', type='float', default=.150,
    help='Frequency to use for source strengths.  Default .150 GHz.')
o.add_option('-j', '--juldate', dest='juldate', type='float',
    help='Julian date used for locating moving sources.')
o.add_option('--nside', dest='nside', type='int',
    help="Manually set NSIDE (possibly degrading map) to a power of 2.")
o.add_option('--nindices', dest='nindices', type='int',default=0,
    help='Number of terms of the spectral index polynomial to save in map.  Default is 0.')
o.add_option('--mscale', dest='mscale', type='float',default=1,
    help='Scaling to apply to data in map.  Useful for changing T into Jy per pixel, and reverse.')
o.add_option('--sscale', dest='sscale', type='float',default=1,
    help='Scaling to apply to sources.  Useful for changing Jy to T.  It may be useful to know that the size of a map pixel is 4*pi / (12*nside**2).')
o.add_option('--iepoch', dest='iepoch', default=e.J2000,
    help='The epoch of coordinates in the input map.  Default J2000.')
o.add_option('--oepoch', dest='oepoch', default=e.J2000, 
    help='The epoch of coordinates in the output map.  Default J2000.')
o.add_option('--isys', dest='isys', default='eq',
    help='Coordinates (eq,ec,ga) of input map.  Default is eq')
o.add_option('--osys', dest='osys', default='eq',
    help='Coordinates (eq,ec,ga) of output map.  Default is eq')
o.add_option('--dtype', dest='dtype', 
    help='Data type (float, double) of output map.  Default is same as input map, or double if no input map is specified.')
opts,args = o.parse_args(sys.argv[1:])

assert(opts.in_map != None or opts.nside != None)
assert(opts.dtype == None or opts.dtype in ['float', 'double'])
if opts.dtype != None:
    if opts.dtype.startswith('fl'): opts.dtype = np.float32
    else: opts.dtype = np.double

if not opts.in_map is None:
    imap = a.map.Map(fromfits=opts.in_map)
    if opts.nside is None and opts.dtype is None:
        print 'Using data from %s' % opts.in_map
        h = imap
    else:
        if opts.dtype is None: opts.dtype = imap.get_dtype()
        if opts.nside is None: opts.nside = imap.nside()
        h = a.map.Map(nside=opts.nside,nindices=opts.nindices,dtype=opts.dtype)
        print 'Importing data from %s' % opts.in_map
        h.from_map(a.map.Map(fromfits=opts.in_map))
    h.map.map *= opts.mscale
    px = np.arange(h.npix())
    m = a.coord.convert_m(opts.osys, opts.isys,
        iepoch=opts.oepoch, oepoch=opts.iepoch)
    x,y,z = np.dot(m,h.px2crd(px, ncrd=3))
    h.set_interpol(True)
    h.map[px] = h.map[x,y,z]
    h.wgt[px] = h.wgt[x,y,z]
    for i in h.ind: i[px] = i[x,y,z]
    h.set_interpol(False)
else:
    if opts.dtype is None: opts.dtype = np.double
    h = a.map.Map(nside=opts.nside, nindices=opts.nindices, dtype=opts.dtype)
    print 'Starting a new map.'
print 'NSIDE:', h.nside()
h.reset_wgt()

if opts.src is None:
    print 'Saving to', opts.map
    h.to_fits(opts.map)
    sys.exit(0)

srclist,cutoff,catalogs = a.scripting.parse_srcs(opts.src, opts.cat)
if not opts.cal is None:
    cat = a.cal.get_catalog(opts.cal, srclist, cutoff, catalogs)
    aa = a.cal.get_aa(opts.cal, .1, opts.freq, 1)
    aa.set_jultime(opts.juldate)
    cat.compute(aa)
else:
    cat = a.src.get_catalog(srclist, cutoff, catalogs)

m = a.coord.convert_m('eq', opts.osys, oepoch=opts.oepoch)
ths,phis = h.px2crd(np.arange(h.npix()), ncrd=2)
ras,decs = phis, np.pi/2 - ths
afreq = np.array([opts.freq])
for srcname in cat:
    src = cat[srcname]
    eq = e.Equatorial(src._ra, src._dec, epoch=e.J2000)
    eq = e.Equatorial(eq, epoch=opts.oepoch)
    ra,dec = eq.get()
    # Account for size of source
    a1,a2,th = src.srcshape
    dras,ddecs = ras - ra, decs - dec
    print '--------------------------------------------------'
    src.update_jys(afreq)
    strength = src.get_jys()[0] * opts.sscale
    print 'Adding', srcname, 'with strength %f Jy' % strength,
    print 'and index', src.index
    print 'Source shape: a1=%f, a2=%f, th=%f' % (a1, a2, th)
    da1 = dras*np.cos(th) - ddecs*np.sin(th)
    da2 = dras*np.sin(th) + ddecs*np.cos(th)
    delta = (da1/a1)**2 + (da2/a2)**2
    px = np.where(delta <= 1)[0]
    if len(px) == 0:
        print 'Treating as point source.'
        eq = a.coord.radec2eq((ra,dec))
        x,y,z = np.dot(m, eq)
        px = h.crd2px(np.array([x]),np.array([y]),np.array([z]))
    str_per_px = strength / len(px)
    print 'Putting %f in each of %d pixels' % (str_per_px, len(px))
    if opts.nindices == 0:
        for p in px:
            curflux = h[p]
            h.put(p,1,str_per_px+curflux)
    else:
        for p in px:
            curflux, i = h[p]
            h.put(p,1,str_per_px+curflux,[src.index])

print 'Saving to', opts.map
h.to_fits(opts.map)




"""
Module for adding polarization information to models.
"""

from aipy import coord,fit,miriad
import numpy as np

#  _   ___     __
# | | | \ \   / /
# | | | |\ \ / /
# | |_| | \ V /
#  \___/   \_/
#

def ijp2blp(i,j,pol):
    return miriad.ij2bl(i,j) * 16 + (pol + 9)

def blp2ijp(blp):
    bl,pol = int(blp) / 16, (blp % 16) - 9
    i,j = miriad.bl2ij(bl)
    return i,j,pol

class UV(miriad.UV):
    def read_pol(self):
        """ Reliably read polarization metadata."""
        return miriad.pol2str[self._rdvr('pol','i')]
    def write_pol(self,pol):
        """Reliably write polarization metadata."""
        try: return self._wrvr('pol','i',miriad.str2pol[pol])
        except(KeyError): 
            print pol,"is not a reasonable polarization value!"
            return

#  _   _ _   _ _ _ _           _____                 _   _                 
# | | | | |_(_) (_) |_ _   _  |  ___|   _ _ __   ___| |_(_) ___  _ __  ___ 
# | | | | __| | | | __| | | | | |_ | | | | '_ \ / __| __| |/ _ \| '_ \/ __|
# | |_| | |_| | | | |_| |_| | |  _|| |_| | | | | (__| |_| | (_) | | | \__ \
#  \___/ \__|_|_|_|\__|\__, | |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
#                      |___/        

def ParAng(ha,dec,lat):
    """
    For any hour angle, declenation in an image, calculate the paralactic angle at that point. Remember to multiply this by 2 when you're
    doing anything with it...
    """
    up = (np.cos(lat)*np.sin(ha))
    down = (np.sin(lat)*np.cos(dec))-(np.cos(lat)*np.sin(dec)*np.cos(ha))
    return np.arctan2(up,down)

#  ____
# | __ )  ___  __ _ _ ___ ___
# |  _ \ / _ \/ _` | '_  `_  \
# | |_) |  __/ (_| | | | | | |
# |____/ \___|\__,_|_| |_| |_|

#     _          _                         
#    / \   _ __ | |_ ___ _ __  _ __   __ _ 
#   / _ \ | '_ \| __/ _ \ '_ \| '_ \ / _` |
#  / ___ \| | | | ||  __/ | | | | | | (_| |
# /_/   \_\_| |_|\__\___|_| |_|_| |_|\__,_|

class Antenna(fit.Antenna):
    '''This is like fit.Antenna, except now it is expected that phsoff, amp,
    bp_r, and bp_i are dicts of {pol:value, ...}.'''
    def _update_phsoff(self):
        self.phsoff = {}
        for pol in self._phsoff: 
            self.phsoff[pol] = np.polyval(self._phsoff[pol], self.beam.afreqs)
    def _update_gain(self):
        self._gain = {}
        for pol in self.bp_r:
            bp = np.polyval(self.bp_r[pol], self.beam.afreqs) + \
                1j*np.polyval(self.bp_i[pol], self.beam.afreqs)
            self._gain[pol] = self.amp[pol] * bp
    def passband(self, conj=False, pol='x'):
        if conj: return np.conjugate(self._gain[pol])
        else: return self._gain[pol]
    def bm_response(self,top,pol='x'):
        """Introduce Stokes parameters in to the definition of the beam."""
        if pol in 'xy':
            return fit.Antenna.bm_response(self,top,pol)
        else:
            assert(pol in 'IQUV')
            if pol in 'IQ': return np.sqrt(0.5*fit.Antenna.bm_response(self,top,pol='x')**2+0.5*fit.Antenna.bm_response(self,top,pol='y')**2)
            if pol in 'UV': return np.sqrt(fit.Antenna.bm_response(self,top,pol='x')*fit.Antenna.bm_response(self,top,pol='y'))
    def get_params(self, prm_list=['*']):
        """Return all fitable parameters in a dictionary."""
        x,y,z = self.pos
        aprms = {'x':x, 'y':y, 'z':z}
        for p in self._phsoff:
            aprms.update({
                'dly_'+p:self._phsoff[p][-2],
                'off_'+p:self._phsoff[p][-1],
                'phsoff_'+p:self._phsoff[p]})
            aprms['bp_r_'+p] = list(self.bp_r[p])
            aprms['bp_i_'+p] = list(self.bp_i[p])
            aprms['amp_'+p] = self.amp[p]
        aprms.update(self.beam.get_params(prm_list))
        prms = {}
        for p in prm_list:
            if p.startswith('*'): return aprms
            try: prms[p] = aprms[p]
            except(KeyError): pass
        return prms
    def set_params(self, prms):
        """Set all parameters from a dictionary."""
        changed = False
        self.beam.set_params(prms)
        try: self.pos[0], changed = prms['x'], True
        except(KeyError): pass
        try: self.pos[1], changed = prms['y'], True
        except(KeyError): pass
        try: self.pos[2], changed = prms['z'], True
        except(KeyError): pass
        for p in self._phsoff:
            try: self._phsoff[p][-2], changed = prms['dly_'+p], True
            except(KeyError): pass
            try: self._phsoff[p][-1], changed = prms['off_'+p], True
            except(KeyError): pass
            try: self._phsoff[p], changed = prms['phsoff_'+p], True
            except(KeyError): pass
            try: self.bp_r[p], changed = prms['bp_r_'+p], True
            except(KeyError): pass
            try: self.bp_i[p], changed = prms['bp_i'+p], True
            except(KeyError): pass
            try: self.amp[p], changed = prms['amp_'+p], True
            except(KeyError): pass
        if changed: self.update()
        return changed

#     _          n                            _                         
#    / \   _ __ | |_ ___ _ __  _ __   __ _   / \   _ __ _ __ __ _ _   _ 
#   / _ \ | '_ \| __/ _ \ '_ \| '_ \ / _` | / _ \ | '__| '__/ _` | | | |
#  / ___ \| | | | ||  __/ | | | | | | (_| |/ ___ \| |  | | | (_| | |_| |
# /_/   \_\_| |_|\__\___|_| |_|_| |_|\__,_/_/   \_\_|  |_|  \__,_|\__, |
#                                                                 |___/ 

class AntennaArray(fit.AntennaArray):
    def set_active_pol(self, pol):
        assert(pol in ('xx','yy','xy','yx','I','Q','U','V')) # Now supports Stokes parameters as well as linear pols
        self.active_pol = pol
    def get_phs_offset(self, i, j):
        pol = self.get_active_pol()
        if pol in ['xx','yy','xy','yx']:
            return self[j].phsoff[pol[-1]] - self[i].phsoff[pol[0]]
        elif pol in ['I','Q','U','V']:
            if pol == 'I':
                return (self[j].phsoff['y'] - self[i].phsoff['y'])+(self[j].phsoff['x'] - self[i].phsoff['x'])
            if pol == 'Q':
                return (self[j].phsoff['y'] - self[i].phsoff['y'])-(self[j].phsoff['x'] - self[i].phsoff['x'])
            if pol == 'U':
                return (self[j].phsoff['y'] - self[i].phsoff['x'])+(self[j].phsoff['x'] - self[i].phsoff['y'])
            if pol == 'V':
                return np.conj((self[j].phsoff['y'] - self[i].phsoff['x'])-(self[j].phsoff['x'] - self[i].phsoff['y']))
    def passband(self, i, j):
        pol = self.get_active_pol()
        return self[j].passband(pol=pol[-1]) * self[i].passband(conj=True, pol=pol[0])



#! /usr/bin/env python
"""
This is a general-purpose script for deconvolving dirty images by a 
corresponding PSF to produce a clean image.
"""

import aipy as a, numpy as np, sys, optparse, ephem, os

o = optparse.OptionParser()
o.set_usage('cl_img.py [options] *.dim.fits *.dbm.fits')
o.set_description(__doc__)
o.add_option('-d', '--deconv', dest='deconv', default='cln',
    help='Attempt to deconvolve the dirty image by the dirty beam using the specified deconvolver (none,mem,lsq,cln,ann).')
o.add_option('-o', '--output', dest='output', default='bim',
    help='Comma delimited list of data to generate FITS files for.  Can be: cim (clean image), cimc (clean image, convolved with central lobe of dirty beam) rim (residual image), or bim (best image = clean + residuals). Default is bim.')
o.add_option('--var', dest='var', type='float', default=.6,
    help='Starting guess for variance in maximum entropy fit (defaults to variance of dirty image.')
o.add_option('--tol', dest='tol', type='float', default=1e-6,
    help='Tolerance for successful deconvolution.  For annealing, interpreted as cooling speed.')
o.add_option('--div', dest='div', action='store_true',
    help='Allow clean to diverge (i.e. allow residual score to increase)')
o.add_option('--minuv', dest='minuv', type='float', default=0.,
    help='Minimum uv length to include (post-gridding)')
o.add_option('--maxuv', dest='maxuv', type='float', default=-1,
    help='Maximum uv length to include (post-gridding)')
o.add_option('-r', '--rewgt', dest='rewgt',  default='natural',
    help='Reweighting to apply to dim/dbm data before cleaning.  Options are: natural, uniform(LEVEL), or radial, where LEVEL is the fractional cutoff for using uniform weighting (recommended range .01 to .1).  Default is natural')
o.add_option('-g','--gain',dest='gain',default=0.1,type='float',help='Loop gain, default=0.1')
o.add_option('--pos_def',dest='pos_def',action='store_true',help="Don't set any negative clean components.")
o.add_option('--maxiter', dest='maxiter', type='int', default=200,
    help='Number of allowable iterations per deconvolve attempt.')
opts, args = o.parse_args(sys.argv[1:])

# Parse command-line options
outputs = opts.output.split(',')
keys = {}
for arg in args: keys[arg[:-len('.dim.fits')]] = None
keys = keys.keys()
keys.sort()

# Define a quick function writing an image to a FITS file
def to_fits(prefix,ftag,i,kwds):
    filename = '%s.%s.fits' % (prefix, ftag)
    print 'Saving data to', filename
    a.img.to_fits(filename, i.astype(np.float32), clobber=True, **kwds)

# Loop through all specified files
for cnt, k in enumerate(keys):
    fdim,fdbm = k+'.dim.fits', k+'.dbm.fits'
    print '-----------------------------------------------------------'
    print '%d / %d' % (cnt + 1, len(keys))
    print 'Deconvolving %s by %s' % (fdim, fdbm)
    if np.all([os.path.exists('%s.%s.fits' % (k,ftag)) for ftag in outputs]):
        print 'All output files exist already... skipping.'
        continue
    dim, kwds = a.img.from_fits(fdim)
    dbm, kwds = a.img.from_fits(fdbm)
    dim = dim.astype(np.float32)
    dbm = dbm.astype(np.float32)
    if np.all(dbm == 0):
        print 'No data in image, so skipping.'
        continue
    print kwds
    size = 1/(kwds['d_ra']*a.img.deg2rad)
    res = size/dim.shape[0]
    im = a.img.Img(size=size, res=res)
    u,v = im.get_uv()
    r = np.sqrt(u**2+v**2)
    dim,dbm = dim.squeeze(), dbm.squeeze()
    DIM = dim.shape[0]
    uvs,bms = np.fft.fft2(dim), np.fft.fft2(dbm)
    if opts.rewgt.startswith('natural'): pass
    else:
        if opts.rewgt.startswith('uniform'): 
            level = float(opts.rewgt.split('(')[-1][:-1])
            abms = np.abs(bms)
            thresh = abms.max() * level
            divisor = abms.clip(thresh, np.Inf)
            uvs /= divisor; bms /= divisor
        elif opts.rewgt.startswith('radial'):
            #x,y = np.indices(dim.shape)
            #x = a.img.recenter(x - DIM/2, (DIM/2,DIM/2))
            #y = a.img.recenter(y - DIM/2, (DIM/2,DIM/2))
            #r = np.sqrt(x**2 + y**2)
            uvs *= r; bms *= r
        elif opts.rewgt.startswith('midrange'):
            x,y = np.indices(dim.shape)
            x = a.img.recenter(x - DIM/2, (DIM/2,DIM/2))
            y = a.img.recenter(y - DIM/2, (DIM/2,DIM/2))
            wgt = x**2 + y**2
            wgt = wgt.astype(np.float32) / (DIM/2)**2
            wgt = wgt * (1-wgt)
            wgt = np.where(wgt < 0, 0, wgt)
            uvs *= wgt; bms *= wgt
        else: raise ValueError('Unrecognized rewgt: %s' % opts.rewgt)
    #mask = np.where(r < opts.minuv, 0, 1)
    #mask = np.where(r > opts.maxuv, 0, mask)
    mask = 1
    if opts.minuv > 0: mask *= 1 - np.exp(-r**2/opts.minuv**2)
    if opts.maxuv > 0: mask *= np.exp(-r**2/opts.maxuv**2)
    dim = np.fft.ifft2(uvs * mask).real
    dbm = np.fft.ifft2(bms * mask).real
    
    dbm = a.img.recenter(dbm, (DIM/2,DIM/2))
    bm_gain = a.img.beam_gain(dbm)
    print 'Gain of dirty beam:', bm_gain
    if opts.deconv == 'mem':
        cim,info = a.deconv.maxent_findvar(dim, dbm, f_var0=opts.var,
            maxiter=opts.maxiter, verbose=True, tol=opts.tol, maxiterok=True)
    elif opts.deconv == 'lsq':
        cim,info = a.deconv.lsq(dim, dbm, 
            maxiter=opts.maxiter, verbose=True, tol=opts.tol)
    elif opts.deconv == 'cln':
        cim,info = a.deconv.clean(dim, dbm, gain=opts.gain, 
            maxiter=opts.maxiter, stop_if_div=not opts.div, 
            verbose=True, tol=opts.tol,pos_def=not opts.pos_def)
    elif opts.deconv == 'ann':
        cim,info = a.deconv.anneal(dim, dbm, maxiter=opts.maxiter, 
            cooling=lambda i,x: opts.tol*(1-np.cos(i/50.))*(x**2), verbose=True)
    else:
        cim,info = np.zeros_like(dim), {'res':dim}
    
    #Fit a 2d Gaussian to the dirty beam and convolve that with the clean components.
    dbm_fit = np.fft.fftshift(dbm)
    DIM = dbm.shape[0]
    lo,hi = (DIM-30)/2,(DIM+30)/2
    dbm_fit = dbm_fit[lo:hi,lo:hi]
    if opts.deconv == 'none': # XXX this is a hack
        rim = info['res']
        bim = rim / bm_gain + cim
    else:
        # This doesn't work for deconv=none
        cbm = a.twodgauss.twodgaussian(a.twodgauss.moments(dbm_fit),shape=dbm.shape)
        cbm = a.img.recenter(cbm,(np.ceil((DIM+dbm_fit.shape[0])/2),np.ceil((DIM+dbm_fit.shape[0])/2)))
        cbm /= np.sum(cbm)

        cimc = np.fft.fftshift(np.fft.ifft2(np.fft.fft2(cim)*np.fft.fft2(cbm))).real

        rim = info['res']

        bim = rim / bm_gain + cimc 
    
    for ftag in ['cim','rim','bim','cimc']:
        print ftag
        if ftag in outputs: to_fits(k, ftag, eval(ftag), kwds)
    




"""
Module for keeping track of physical constants.  All constants in cgs units.
See description() for a dictionary of constants and their descriptions.
"""

pi = 3.141592653589793      # Pi
c = 2.99792458e10           # Speed of Light
G = 6.6726e-8               # Newton's Gravitational Constant
h = 6.6260755e-27           # Planck constant
e = 4.8032068e-10           # Elementary Charge
m_e = 9.1093898e-28         # Electron Mass
m_p = 1.6726231e-24         # Proton Mass
k = 1.380658e-16            # Boltzmann constant
s = 5.67051e-5              # Stefan-Boltzmann constant
au = 1.496e13               # cm in 1 AU
r_sun = 6.96e10             # radius of sun (cm)
m_sun = 1.98892e33          # mass of sun (g)
ev = 1.60217733e-12         # Electron Volts
pc = 3.09e18                # cm in 1 pc
km = 1e5                    # cm in 1 km
s_per_day = 24.*60*60       # seconds in a solar day
s_per_yr = 3.1556926e7      # seconds in a year
sidereal_day = 86164.0908   # seconds
len_ns = c * 1e-9           # length of a nanosecond in cm
deg = pi / 180.             # degrees in radians
sq_deg = 3.04617e-4         # square degree in steradians
arcmin = deg / 60.          # arcminute in radians
arcsec = arcmin / 60.       # arcsecond in radians
ft = 30.48006096012         # length of a foot in cm

def description():
    return {
        'pi'          : "Pi",
        'c'           : "Speed of Light",
        'G'           : "Newton's Gravitational Constant",
        'h'           : "Planck constant",
        'e'           : "Elementary Charge",
        'm_e'         : "Electron Mass",
        'm_p'         : "Proton Mass",
        'k'           : "Boltzmann constant",
        's'           : "Stefan-Boltzmann constant",
        'au'          : "cm in 1 AU",
        'r_sun'       : "radius of sun (cm)",
        'm_sun'       : "mass of sun (g)",
        'ev'          : "Electron Volts",
        'pc'          : "cm in 1 pc",
        'km'          : "cm in 1 km",
        's_per_day'   : "seconds in a solar day",
        's_per_yr'    : "seconds in a year",
        'sidereal_day': "seconds in a sidereal day",
        'len_ns'      : "length of a nanosecond in cm",
        'deg'         : "degrees in radians",
        'sq_deg'      : "square degree in steradians",
        'arcmin'      : "arcminute in radians",
        'arcsec'      : "arcsecond in radians",
        'ft'          : "length of a foot in cm",
    }



import aipy as a, numpy as np

specials = ['Sun', 'Moon', 'Mercury', 'Venus', 'Mars', 'Jupiter',
    'Saturn', 'Uranus', 'Neptune']
src_data = {
    #             RA          DEC         FLUX  FREQ, INDEX ANGSIZE
    'Moon'  : (None        , None,           0, .150,  0., 0.),
    'Mercury': (None        , None,          0, .150,  0., 0.),
    'Venus':   (None        , None,          0, .150,  0., 0.),
    'Mars':    (None        , None,          0, .150,  0., 0.),
    'Jupiter': (None        , None,          0, .150,  0., 0.),
    'Saturn' : (None        , None,          0, .150,  0., 0.),
    'Uranus' : (None        , None,          0, .150,  0., 0.),
    'Neptune': (None        , None,          0, .150,  0., 0.),
    'Sun'  : (None, None, 57000, .150,  2.00, 4.6e-3),
    'for':   ('03:22:41.7', '-37:12:30',   170, .150, -0.8 , 0.),
    'pic'  : ('05:19:49.7', '-45:46:45',   452, .160, -0.8 , 0.),
    'hyd'  : ('09:18:05.7', '-12:05:44',  1860, .160, -2.30, 0.),
    'cen'  : ('13:25:27.6', '-43:01:09', 5144, .150, -0.65, 0.), # From Israel 1998
    'her' : ('16:51:08.15', '4:59:33.3', 300.0, 0.159, -1, 0.000669),
    'sgr'  : ('17:45:40.0', '-29:00:28',  121, .160, -4.21, 0.),
    # Fluxes from Miriad:
    'crab' : ('05:34:32.0', '+22:00:52',  1838, .150, -0.30, 0.),
    'vir'  : ('12:30:49.4', '+12:23:28',   1446, .150, -0.86, 0.),
    'cyg'  : ('19:59:28.3', '+40:44:02', 10900, .150, -0.69, (0.,0.,0.)),
    'cas'  : ('23:23:27.94', '+58:48:42.4',  9160.0, 0.150, -0.73, 0.000000),
}

_misccat = None

def get_srcs(srcs=None, cutoff=None):
    global _misccat
    if _misccat is None:
        _misccat = a.fit.SrcCatalog()
        srclist = []
        for s in src_data:
            ra, dec, st, mfreq, index, srcshape = src_data[s]
            try: len(srcshape)
            except(TypeError): srcshape = (srcshape, srcshape, 0.)
            if s in specials:
                srclist.append(a.fit.RadioSpecial(s, st, mfreq=mfreq, 
                    index=index, srcshape=srcshape))
            else:
                srclist.append(a.fit.RadioFixedBody(ra, dec, jys=st, mfreq=mfreq, 
                    index=index, name=s, srcshape=srcshape))
        _misccat.add_srcs(srclist)
    if srcs is None:
        if cutoff is None: srcs = _misccat.keys()
        else:
            cut, fq = cutoff
            fq = np.array([fq])
            for s in _misccat.keys(): _misccat[s].update_jys(fq)
            srcs = [s for s in _misccat.keys() if _misccat[s].jys[0] > cut]

    srclist = []
    for s in srcs:
        try: srclist.append(_misccat[s])
        except(KeyError): pass
    return srclist




"""
Module for fitting 2d Gaussians to things.
"""

import numpy as np

def moments(data):
    total = np.abs(data).sum()
    Y,X = np.indices(data.shape)
    y = np.argmax((X*np.abs(data)).sum(axis=1)/total)
    x = np.argmax((Y*np.abs(data)).sum(axis=0)/total)
    col = data[int(y),:]
    row = data[:,int(x)]
    #First moment.
    width_x = np.sqrt(np.abs((np.arange(col.size)-y)*col).sum()/np.abs(col).sum())
    width_y = np.sqrt(np.abs((np.arange(row.size)-x)*row).sum()/np.abs(row).sum())
    width = ( width_x + width_y ) / 2.
    height = np.median(data.ravel())
    amplitude = data.max() - height
    mylist = [amplitude,x,y]
    if np.isnan(width_y) or np.isnan(width_x) or np.isnan(height) or np.isnan(amplitude):
        raise ValueError("Somehthing is nan")
    mylist = [height] + mylist
    mylist = mylist + [width_x,width_y]
    return mylist

def twodgaussian(inpars,shape=None):
    inpars_old = inpars
    inpars = list(inpars)
    height = inpars.pop(0)
    height = float(height)
    amplitude,center_y,center_x = inpars.pop(0),inpars.pop(0),inpars.pop(0)
    amplitude = float(amplitude)
    center_x = float(center_x)
    center_y = float(center_y)
    width_x,width_y = inpars.pop(0),inpars.pop(0)
    rcen_x = center_x
    rcen_y = center_y
    if len(inpars) > 0:
        raise ValueError("There are still input parameters:" + str(inpars) + \
                        " and you've input: " + str(inpars_old) + \
                        " circle=%d, rotate=%d, vheight=%d" % (circle,rotate,vheight) )
    def rotgauss(x,y):
        xp = x
        yp = y
        g = amplitude*np.exp( 
            -(((rcen_x-xp)/width_x)**2 + 
            ((rcen_y-yp)/width_y)**2)/2.)
        return g
    if shape is not None:
        return rotgauss(*np.indices(shape))
    else:
        return rotgauss



#! /usr/bin/env python
import aipy as a, numpy as np
import optparse, sys, os

o = optparse.OptionParser()
a.scripting.add_standard_options(o, ant=True)
o.set_usage('xrfi_simple.py [options] *.uv')
o.set_description(__doc__)
o.add_option('-c', '--chan', dest='chan',
    help='Manually flag channels before xrfi processing.  Options are "<chan1 #>,..." (a list of channels), or "<chan1 #>_<chan2 #>" (a range of channels).  Default is None.')
o.add_option('-n', '--nsig', dest='nsig', default=2., type='float',
    help='Number of standard deviations above mean to flag if neither --dt nor --df are specified.  Default 2.')
o.add_option('--df', dest='df', type='float', 
    help='Number of standard deviations above mean to flag, after taking derivative of frequency axis')
o.add_option('--dt', dest='dt', type='float',
    help='Number of standard deviations above mean to flag, after taking derivative of time axis')
o.add_option('--combine', dest='combine', action='store_true',
    help='Use the same mask for all baselines/pols (and use thresh to decide how many concidences it takes to flag all data.')
o.add_option('--to_npz', 
    help='Instead of applying mask to data, store it as npz of this name.  May only be used along with --combine.')
o.add_option('--from_npz', 
    help='Apply mask to data from this npz file (generated with --to_npz).  May only be used along with --combine.')
o.add_option('-t', '--thresh', dest='thresh', default=1, type='int',
    help='Number of flagging coincidences (baselines/pols) required to flag a time/chan.')
opts,args = o.parse_args(sys.argv[1:])

# Parse command-line options
uv = a.miriad.UV(args[0])
if not opts.chan is None:
    chans = a.scripting.parse_chans(opts.chan, uv['nchan'])
else:
    chans = []
    opts.chan = 'None'
del(uv)
if opts.to_npz or opts.from_npz: assert(opts.combine)


for uvfile in args:
    uvofile = uvfile+'R'
    print uvfile,'->',uvofile
    if os.path.exists(uvofile):
        print uvofile, 'exists, skipping.'
        continue
    if opts.from_npz:
        print '    Reading flags from', opts.from_npz
        m = np.load(opts.from_npz)
        mask = {'xx':{257:{}}} # Just use dummy values here to mimic structure of mask dictionary
        for cnt,t in enumerate(m['times']):
            mask['xx'][257][t] = m[str(cnt)]
    else:
        uvi = a.miriad.UV(uvfile)
        a.scripting.uv_selector(uvi, opts.ant)
        # Gather all data and each time step
        data,mask,times = {}, {}, []
        for (uvw,t,(i,j)), d, f in uvi.all(raw=True):
            if len(times) == 0 or times[-1] != t: times.append(t)
            bl = a.miriad.ij2bl(i,j)
            pol = uvi['pol']
            if not pol in data:
                data[pol] = {}
                mask[pol] = {}
            if not bl in data[pol]:
                data[pol][bl] = {}
                mask[pol][bl] = {}
            # Manually flag channels
            f[chans] = 1
            mask[pol][bl][t] = f
            data[pol][bl][t] = d

        # Generate statistical mask
        for pol in data:
          for bl in data[pol]:
            i, j = a.miriad.bl2ij(bl)
            data_times = data[pol][bl].keys()
            data_times.sort()
            d = np.array([data[pol][bl][t] for t in data_times])
            m = np.array([mask[pol][bl][t] for t in data_times])
            if opts.df != None:
                ddf = d[:,1:-1] - .5 * (d[:,:-2] + d[:,2:])
                ddf2 = np.abs(ddf)**2
                sig = np.sqrt(np.median(ddf2, axis=1))
                sig.shape = (sig.size,1)
                m[:,0] |= True; m[:,-1] |= True
                m[:,1:-1] |= np.where(ddf2/sig**2 > opts.df**2, True, False)
            if opts.dt != None:
                ddt = d[1:-1,:] - .5 * (d[:-2,:] + d[2:,:])
                ddt2 = np.abs(ddt)**2
                sig = np.sqrt(np.median(ddt2, axis=0))
                sig.shape = (1,sig.size)
                m[0,:] |= True; m[-1,:] |= True
                m[1:-1,:] |= np.where(ddt2/sig**2 > opts.dt**2, True, False)
            if opts.df == None and opts.dt == None:
                ad = np.abs(d)
                med = np.median(ad)
                sig = np.sqrt(np.median(np.abs(ad-med)**2))
                m |= np.where(ad > med + opts.nsig * sig, True, False)
            for i, t in enumerate(data_times): mask[pol][bl][t] |= m[i]
        if opts.combine:
            new_mask = {}
            for pol in mask:
              for bl in mask[pol]:
                for t in mask[pol][bl]:
                    new_mask[t] = new_mask.get(t,0)+mask[pol][bl][t].astype(np.int)
            for t in new_mask:
                m = np.where(new_mask[t] >= opts.thresh, 1, 0)
                for pol in mask:
                  for bl in mask[pol]:
                    mask[pol][bl][t] = m
        del(uvi)

    if opts.to_npz:
        print '    Writing flags to', opts.to_npz
        m = {}
        _m = mask.values()[0].values()[0]
        times = np.array(_m.keys())
        for cnt,t in enumerate(times): m[str(cnt)] = _m[t]
        m['times'] = times
        np.savez(opts.to_npz, **m)
    else:
        # Generate a pipe for applying the mask to data as it comes in.
        def rfi_mfunc(uv, preamble, data, flags):
            uvw, t, (i,j) = preamble
            bl = a.miriad.ij2bl(i,j)
            if opts.combine:
                try: m = mask.values()[0].values()[0][t]
                except(KeyError): m = np.ones_like(flags) # default to flagging
            else: m = mask[uv['pol']][bl][t]
            return preamble, np.where(m, 0, data), m

        uvi = a.miriad.UV(uvfile)
        uvo = a.miriad.UV(uvofile, status='new')
        uvo.init_from_uv(uvi)
        uvo.pipe(uvi, mfunc=rfi_mfunc, raw=True, append2hist=' '.join(sys.argv)+'\n')



